from __future__ import annotations

import asyncio
import logging
import os
import re
from concurrent.futures import Future, ThreadPoolExecutor, wait
from datetime import date, datetime, timedelta
from pathlib import Path

import typer
from ask_shell import confirm, new_task, print_to_live, run_and_wait, select_list
from model_lib import Entity, Event, copy_and_validate
from pydantic import Field, ValidationError, field_validator, model_validator
from pydantic_core import Url
from rich.markdown import Markdown
from zero_3rdparty import file_utils
from zero_3rdparty.datetime_utils import utc_now
from zero_3rdparty.str_utils import ensure_suffix

from atlas_init.cli_helper.run import add_to_clipboard
from atlas_init.cli_tf.github_logs import (
    GH_TOKEN_ENV_NAME,
    download_job_safely,
    is_test_job,
    tf_repo,
)
from atlas_init.cli_tf.go_test_run import GoTestRun, GoTestStatus, parse_tests
from atlas_init.cli_tf.go_test_summary import (
    DailyReportIn,
    DailyReportOut,
    ErrorRowColumns,
    MonthlyReportIn,
    RunHistoryFilter,
    TFCITestOutput,
    TestRow,
    create_daily_report,
    create_monthly_report,
)
from atlas_init.cli_tf.go_test_tf_error import (
    DetailsInfo,
    ErrorClassAuthor,
    GoTestError,
    GoTestErrorClass,
    GoTestErrorClassification,
    parse_error_details,
)
from atlas_init.cli_tf.mock_tf_log import resolve_admin_api_path
from atlas_init.crud.mongo_dao import (
    TFResources,
    init_mongo_dao,
    read_tf_resources,
)
from atlas_init.html_out.md_export import MonthlyReportPaths, export_ci_tests_markdown_to_html
from atlas_init.repos.go_sdk import ApiSpecPaths, parse_api_spec_paths
from atlas_init.repos.path import Repo, current_repo_path
from atlas_init.settings.env_vars import AtlasInitSettings, init_settings

logger = logging.getLogger(__name__)


class TFCITestInput(Event):
    settings: AtlasInitSettings = Field(default_factory=init_settings)
    repo_path: Path = Field(default_factory=lambda: current_repo_path(Repo.TF))
    test_group_name: str = ""
    max_days_ago: int = 1
    branch: str = "master"
    workflow_file_stems: set[str] = Field(default_factory=lambda: set(_TEST_STEMS))
    names: set[str] = Field(default_factory=set)
    skip_log_download: bool = False
    skip_error_parsing: bool = False
    summary_name: str = ""
    report_date: datetime = Field(default_factory=utc_now)

    @field_validator("report_date", mode="before")
    def support_today(cls, value: str | datetime) -> datetime | str:
        return utc_now() if isinstance(value, str) and value == "today" else value

    @model_validator(mode="after")
    def set_workflow_file_stems(self) -> TFCITestInput:
        if not self.workflow_file_stems:
            self.workflow_file_stems = set(_TEST_STEMS)
        return self


def ci_tests(
    test_group_name: str = typer.Option("", "-g"),
    max_days_ago: int = typer.Option(
        1, "-d", "--days", help="number of days to look back, Github only store logs for 30 days."
    ),
    branch: str = typer.Option("master", "-b", "--branch"),
    workflow_file_stems: str = typer.Option("test-suite,terraform-compatibility-matrix", "-w", "--workflow"),
    names: str = typer.Option(
        "",
        "-n",
        "--test-names",
        help="comma separated list of test names to filter, e.g., TestAccCloudProviderAccessAuthorizationAzure_basic,TestAccBackupSnapshotExportBucket_basicAzure",
    ),
    summary_name: str = typer.Option(
        ...,
        "-s",
        "--summary",
        help="the name of the summary directory to store detailed test results",
        default_factory=lambda: utc_now().strftime("%Y-%m-%d"),
    ),
    summary_env_name: str = typer.Option("", "--env", help="filter summary based on tests/errors only in dev/qa"),
    skip_log_download: bool = typer.Option(False, "-sld", "--skip-log-download", help="skip downloading logs"),
    skip_error_parsing: bool = typer.Option(
        False, "-sep", "--skip-error-parsing", help="skip parsing errors, usually together with --skip-log-download"
    ),
    skip_daily: bool = typer.Option(False, "-sd", "--skip-daily", help="skip daily report"),
    skip_monthly: bool = typer.Option(False, "-sm", "--skip-monthly", help="skip monthly report"),
    ask_to_open: bool = typer.Option(False, "--open", "--ask-to-open", help="ask to open the reports"),
    copy_to_clipboard: bool = typer.Option(
        False,
        "--copy",
        help="copy the summary to clipboard",
    ),
    report_date: str = typer.Option(
        "today",
        "-rd",
        "--report-day",
        help="the day to generate the report for, defaults to today, format=YYYY-MM-DD",
    ),
):
    names_set: set[str] = set()
    if names:
        names_set.update(names.split(","))
        logger.info(f"filtering tests by names: {names_set} (todo: support this)")
    if test_group_name:
        logger.warning(f"test_group_name is not supported yet: {test_group_name}")
    event = TFCITestInput(
        test_group_name=test_group_name,
        max_days_ago=max_days_ago,
        report_date=report_date,  # type: ignore
        branch=branch,
        workflow_file_stems=set(workflow_file_stems.split(",")),
        names=names_set,
        summary_name=summary_name,
        skip_log_download=skip_log_download,
        skip_error_parsing=skip_error_parsing,
    )
    history_filter = RunHistoryFilter(
        run_history_start=event.report_date - timedelta(days=event.max_days_ago),
        run_history_end=event.report_date,
        env_filter=[summary_env_name] if summary_env_name else [],
    )
    settings = event.settings
    report_paths = MonthlyReportPaths.from_settings(settings, summary_name)
    if skip_daily:
        logger.info("skipping daily report")
    else:
        run_daily_report(event, settings, history_filter, copy_to_clipboard, report_paths)
    if summary_name.lower() != "none":
        monthly_input = MonthlyReportIn(
            name=summary_name,
            branch=event.branch,
            history_filter=history_filter,
            report_paths=report_paths,
        )
        if skip_monthly:
            logger.info("skipping monthly report")
        else:
            generate_monthly_summary(settings, monthly_input, ask_to_open)
    export_ci_tests_markdown_to_html(settings, report_paths)


def run_daily_report(
    event: TFCITestInput,
    settings: AtlasInitSettings,
    history_filter: RunHistoryFilter,
    copy_to_clipboard: bool,
    report_paths: MonthlyReportPaths,
) -> DailyReportOut:
    out = asyncio.run(ci_tests_pipeline(event))
    manual_classification(out.classified_errors, settings)
    summary_name = event.summary_name

    def add_md_link(row: TestRow, row_dict: dict[str, str]) -> dict[str, str]:
        if not summary_name:
            return row_dict
        old_details = row_dict[ErrorRowColumns.DETAILS_SUMMARY]
        old_details = old_details or "Test History"
        row_dict[ErrorRowColumns.DETAILS_SUMMARY] = (
            f"[{old_details}]({settings.github_ci_summary_details_rel_path(summary_name, row.full_name)})"
        )
        return row_dict

    daily_in = DailyReportIn(
        report_date=event.report_date,
        history_filter=history_filter,
        row_modifier=add_md_link,
    )
    daily_out = create_daily_report(out, settings, daily_in)
    print_to_live(Markdown(daily_out.summary_md))
    if copy_to_clipboard:
        add_to_clipboard(daily_out.summary_md, logger=logger)
    file_utils.ensure_parents_write_text(report_paths.daily_path, daily_out.summary_md)
    return daily_out


def generate_monthly_summary(
    settings: AtlasInitSettings, monthly_input: MonthlyReportIn, ask_to_open: bool = False
) -> None:
    monthly_out = create_monthly_report(
        settings,
        monthly_input,
    )
    paths = monthly_input.report_paths
    summary_path = paths.summary_path
    file_utils.ensure_parents_write_text(summary_path, monthly_out.summary_md)
    logger.info(f"summary written to {summary_path}")
    details_dir = paths.details_dir
    file_utils.clean_dir(details_dir, recreate=True)
    logger.info(f"Writing details to {details_dir}")
    for name, details_md in monthly_out.test_details_md.items():
        details_path = details_dir / ensure_suffix(name, ".md")
        file_utils.ensure_parents_write_text(details_path, details_md)
    monthly_error_only_out = create_monthly_report(
        settings,
        event=copy_and_validate(
            monthly_input,
            skip_rows=[MonthlyReportIn.skip_if_no_failures],
            existing_details_md=monthly_out.test_details_md,
        ),
    )
    error_only_path = paths.error_only_path
    file_utils.ensure_parents_write_text(error_only_path, monthly_error_only_out.summary_md)
    logger.info(f"error-only summary written to {error_only_path}")
    if ask_to_open and confirm(f"do you want to open the summary file? {summary_path}", default=False):
        run_and_wait(f'code "{summary_path}"')
    if ask_to_open and confirm(f"do you want to open the error-only summary file? {error_only_path}", default=False):
        run_and_wait(f'code "{error_only_path}"')
    return None


async def ci_tests_pipeline(event: TFCITestInput) -> TFCITestOutput:
    repo_path = event.repo_path
    branch = event.branch
    settings = event.settings
    download_input = DownloadJobLogsInput(
        branch=branch,
        max_days_ago=event.max_days_ago,
        end_date=event.report_date,
        workflow_file_stems=event.workflow_file_stems,
        repo_path=repo_path,
    )
    dao = await init_mongo_dao(settings)
    if event.skip_log_download:
        logger.info("skipping log download, reading existing instead")
        log_paths = []
    else:
        log_paths = download_logs(download_input, settings)
        resources = read_tf_resources(settings, repo_path, branch)
        with new_task(f"parse job logs from {len(log_paths)} files"):
            parse_job_output = parse_job_tf_test_logs(
                ParseJobLogsInput(
                    settings=settings,
                    log_paths=log_paths,
                    resources=resources,
                    branch=branch,
                )
            )
        await dao.store_tf_test_runs(parse_job_output.test_runs)
    report_date = event.report_date
    with new_task(f"reading test runs from storage for {report_date.date().isoformat()}"):
        report_tests = await dao.read_tf_tests_for_day(event.branch, report_date)
    with new_task("parsing test errors"):
        report_errors = parse_test_errors(report_tests)
    with new_task("classifying errors"):
        error_run_ids = [error.run_id for error in report_errors]
        existing_classifications = await dao.read_error_classifications(error_run_ids)
        classified_errors = classify_errors(existing_classifications, report_errors)
    return TFCITestOutput(
        log_paths=log_paths, found_tests=report_tests, classified_errors=classified_errors, found_errors=report_errors
    )


def parse_test_errors(found_tests: list[GoTestRun]) -> list[GoTestError]:
    admin_api_path = resolve_admin_api_path(sdk_branch="main")
    spec_paths = ApiSpecPaths(method_paths=parse_api_spec_paths(admin_api_path))
    error_tests = [test for test in found_tests if test.is_failure]
    test_errors: list[GoTestError] = []
    for test in error_tests:
        test_error_input = ParseTestErrorInput(test=test, api_spec_paths=spec_paths)
        test_errors.append(parse_test_error(test_error_input))
    return test_errors


def classify_errors(
    existing: dict[str, GoTestErrorClassification], errors: list[GoTestError]
) -> list[GoTestErrorClassification]:
    needs_classification: list[GoTestError] = []
    classified_errors: list[GoTestErrorClassification] = []
    for error in errors:
        if prev_classification := existing.get(error.run_id):
            logger.info(f"found existing classification{error.run_name}: {prev_classification}")
            classified_errors.append(prev_classification)
            continue
        if auto_class := GoTestErrorClass.auto_classification(error.run.output_lines_str):
            logger.info(f"auto class for {error.run_name}: {auto_class}")
            classified_errors.append(
                GoTestErrorClassification(
                    error_class=auto_class,
                    confidence=1.0,
                    details=error.details,
                    test_output=error.run.output_lines_str,
                    run_id=error.run_id,
                    author=ErrorClassAuthor.AUTO,
                    test_name=error.run_name,
                )
            )
        else:
            needs_classification.append(error)
    return classified_errors + add_llm_classifications(needs_classification)


def manual_classification(
    classifications: list[GoTestErrorClassification], settings: AtlasInitSettings, confidence_threshold: float = 1.0
):
    needs_classification = [cls for cls in classifications if cls.needs_classification(confidence_threshold)]
    with new_task("Manual Classification", total=len(needs_classification) + 1, log_updates=True) as task:
        asyncio.run(classify(needs_classification, settings, task))


async def classify(
    needs_classification: list[GoTestErrorClassification], settings: AtlasInitSettings, task: new_task
) -> None:
    dao = await init_mongo_dao(settings)

    async def add_classification(
        cls: GoTestErrorClassification, new_class: GoTestErrorClass, new_author: ErrorClassAuthor, confidence: float
    ):
        cls.error_class = new_class
        cls.author = new_author
        cls.confidence = confidence
        is_new = await dao.add_classification(cls)
        if not is_new:
            logger.debug("replaced existing class")

    for cls in needs_classification:
        task.update(advance=1)
        similars = await dao.read_similar_error_classifications(cls.details, author_filter=ErrorClassAuthor.HUMAN)
        if (existing := similars.get(cls.run_id)) and not existing.needs_classification():
            logger.debug(f"found existing classification: {existing}")
            continue
        if similars and len({similar.error_class for similar in similars.values()}) == 1:
            _, similar = similars.popitem()
            if not similar.needs_classification(0.0):
                logger.info(f"using similar classification: {similar}")
                await add_classification(cls, similar.error_class, ErrorClassAuthor.SIMILAR, 1.0)
                continue
        test = await dao.read_tf_test_run(cls.run_id)
        if new_class := ask_user_to_classify_error(cls, test):
            await add_classification(cls, new_class, ErrorClassAuthor.HUMAN, 1.0)
        elif confirm("do you want to stop classifying errors?", default=True):
            logger.info("stopping classification")
            return


def add_llm_classifications(needs_classification_errors: list[GoTestError]) -> list[GoTestErrorClassification]:
    """Todo: Use LLM and support reading existing classifications, for example matching on the details"""
    return [
        GoTestErrorClassification(
            ts=utc_now(),
            error_class=GoTestErrorClass.UNKNOWN,
            confidence=0.0,
            details=error.details,
            test_output=error.run.output_lines_str,
            run_id=error.run_id,
            author=ErrorClassAuthor.LLM,
            test_name=error.run_name,
        )
        for error in needs_classification_errors
    ]


class DownloadJobLogsInput(Entity):
    branch: str = "master"
    workflow_file_stems: set[str] = Field(default_factory=lambda: set(_TEST_STEMS))
    max_days_ago: int = 1
    end_date: datetime = Field(default_factory=utc_now)
    repo_path: Path

    @property
    def start_date(self) -> datetime:
        return self.end_date - timedelta(days=self.max_days_ago)

    @model_validator(mode="after")
    def check_max_days_ago(self) -> DownloadJobLogsInput:
        if self.max_days_ago > 90:
            logger.warning(f"max_days_ago for {type(self).__name__} must be less than or equal to 90, setting to 90")
            self.max_days_ago = 90
        return self


def download_logs(event: DownloadJobLogsInput, settings: AtlasInitSettings) -> list[Path]:
    token = run_and_wait("gh auth token", cwd=event.repo_path).stdout
    assert token, "expected token, but got empty string"
    os.environ[GH_TOKEN_ENV_NAME] = token
    end_test_date = event.end_date
    start_test_date = event.start_date
    log_paths = []
    with new_task(
        f"downloading logs for {event.branch} from {start_test_date.date()} to {end_test_date.date()}",
        total=(end_test_date - start_test_date).days,
    ) as task:
        while start_test_date <= end_test_date:
            event_out = download_gh_job_logs(
                settings,
                DownloadJobRunsInput(branch=event.branch, run_date=start_test_date.date()),
            )
            log_paths.extend(event_out.log_paths)
            if errors := event_out.log_errors():
                logger.warning(errors)
            start_test_date += timedelta(days=1)
            task.update(advance=1)
    return log_paths


_TEST_STEMS = {
    "test-suite",
    "terraform-compatibility-matrix",
    "acceptance-tests",
}


class DownloadJobRunsInput(Event):
    branch: str = "master"
    run_date: date
    workflow_file_stems: set[str] = Field(default_factory=lambda: set(_TEST_STEMS))
    worker_count: int = 10
    max_wait_seconds: int = 300


class DownloadJobRunsOutput(Entity):
    job_download_timeouts: int = 0
    job_download_empty: int = 0
    job_download_errors: int = 0
    log_paths: list[Path] = Field(default_factory=list)

    def log_errors(self) -> str:
        if not (self.job_download_timeouts or self.job_download_empty or self.job_download_errors):
            return ""
        return f"job_download_timeouts: {self.job_download_timeouts}, job_download_empty: {self.job_download_empty}, job_download_errors: {self.job_download_errors}"


def created_on_day(create: date) -> str:
    date_fmt = year_month_day(create)
    return f"{date_fmt}T00:00:00Z..{date_fmt}T23:59:59Z"


def year_month_day(create: date) -> str:
    return create.strftime("%Y-%m-%d")


def download_gh_job_logs(settings: AtlasInitSettings, event: DownloadJobRunsInput) -> DownloadJobRunsOutput:
    repository = tf_repo()
    branch = event.branch
    futures: list[Future[Path | None]] = []
    run_date = event.run_date
    out = DownloadJobRunsOutput()
    with ThreadPoolExecutor(max_workers=event.worker_count) as pool:
        for workflow in repository.get_workflow_runs(
            created=created_on_day(run_date),
            branch=branch,  # type: ignore
        ):
            workflow_stem = Path(workflow.path).stem
            if workflow_stem not in event.workflow_file_stems:
                continue
            workflow_dir = (
                settings.github_ci_run_logs / branch / year_month_day(run_date) / f"{workflow.id}_{workflow_stem}"
            )
            logger.info(f"workflow dir for {workflow_stem} @ {workflow.created_at.isoformat()}: {workflow_dir}")
            if workflow_dir.exists():
                paths = list(workflow_dir.rglob("*.log"))
                logger.info(f"found {len(paths)} logs in existing workflow dir: {workflow_dir}")
                out.log_paths.extend(paths)
                continue
            futures.extend(
                pool.submit(download_job_safely, workflow_dir, job)
                for job in workflow.jobs("all")
                if is_test_job(job.name)
            )
        done, not_done = wait(futures, timeout=event.max_wait_seconds)
        out.job_download_timeouts = len(not_done)
        for future in done:
            try:
                if log_path := future.result():
                    out.log_paths.append(log_path)
                else:
                    out.job_download_empty += 1
            except Exception as e:
                logger.error(f"failed to download job logs: {e}")
                out.job_download_errors += 1
    return out


class ParseJobLogsInput(Event):
    settings: AtlasInitSettings
    log_paths: list[Path]
    resources: TFResources
    branch: str


class ParseJobLogsOutput(Event):
    test_runs: list[GoTestRun] = Field(default_factory=list)

    def tests_with_status(self, status: GoTestStatus) -> list[GoTestRun]:
        return [test for test in self.test_runs if test.status == status]


def parse_job_tf_test_logs(
    event: ParseJobLogsInput,
) -> ParseJobLogsOutput:
    out = ParseJobLogsOutput()
    for log_path in event.log_paths:
        log_text = log_path.read_text()
        env = find_env_of_mongodb_base_url(log_text)
        try:
            result = parse_tests(log_text.splitlines())
        except ValidationError as e:
            logger.warning(f"failed to parse tests from {log_path}: {e}")
            continue
        for test in result:
            test.log_path = log_path
            test.env = env or "unknown"
            test.resources = event.resources.find_test_resources(test)
            test.branch = event.branch
        out.test_runs.extend(result)
    return out


def find_env_of_mongodb_base_url(log_text: str) -> str:
    for match in re.finditer(r"MONGODB_ATLAS_BASE_URL: (.*)$", log_text, re.MULTILINE):
        full_url = match.group(1)
        parsed = BaseURLEnvironment(url=Url(full_url))
        return parsed.env
    return ""


class BaseURLEnvironment(Entity):
    """
    >>> BaseURLEnvironment(url="https://cloud-dev.mongodb.com/").env
    'dev'
    """

    url: Url
    env: str = ""

    @model_validator(mode="after")
    def set_env(self) -> BaseURLEnvironment:
        host = self.url.host
        assert host, f"host not found in url: {self.url}"
        cloud_env = host.split(".")[0]
        self.env = cloud_env.removeprefix("cloud-")
        return self


class ParseTestErrorInput(Event):
    test: GoTestRun
    api_spec_paths: ApiSpecPaths | None = None


def parse_test_error(event: ParseTestErrorInput) -> GoTestError:
    run = event.test
    assert run.is_failure, f"test is not failed: {run.name}"
    details = parse_error_details(run)
    info = DetailsInfo(run=run, paths=event.api_spec_paths)
    details.add_info_fields(info)
    return GoTestError(details=details, run=run)


def ask_user_to_classify_error(cls: GoTestErrorClassification, test: GoTestRun) -> GoTestErrorClass | None:
    details = cls.details
    try:
        print_to_live(test.output_lines_str)
        print_to_live(f"error details: {details}")
        return select_list(
            f"choose classification for test='{test.name_with_package}' in {test.env}",
            choices=list(GoTestErrorClass),
            default=cls.error_class,
        )  # type: ignore
    except KeyboardInterrupt:
        return None



from __future__ import annotations

import logging
from collections import defaultdict
from pathlib import Path
from typing import Iterable, NamedTuple

import pydot
from ask_shell import ShellError, new_task, run_and_wait
from ask_shell.run_pool import run_pool
from ask_shell._run import stop_runs_and_pool
from model_lib import Entity, dump
from pydantic import BaseModel, Field
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed
from typer import Typer
from zero_3rdparty.file_utils import ensure_parents_write_text
from zero_3rdparty.iter_utils import flat_map

from atlas_init.settings.rich_utils import configure_logging
from atlas_init.tf_ext.args import REPO_PATH_ARG, SKIP_EXAMPLES_DIRS_OPTION
from atlas_init.tf_ext.constants import ATLAS_PROVIDER_NAME
from atlas_init.tf_ext.paths import find_variable_resource_type_usages, find_variables, get_example_directories
from atlas_init.tf_ext.settings import TfDepSettings

logger = logging.getLogger(__name__)
v2_grand_parent_dirs = {
    "module_maintainer",
    "module_user",
    "migrate_cluster_to_advanced_cluster",
    "mongodbatlas_backup_compliance_policy",
}
v2_parent_dir = {"cluster_with_schedule"}
MODULE_PREFIX = "module."
DATA_PREFIX = "data."
VARIABLE_RESOURCE_MAPPING: dict[str, str] = {
    "org_id": "mongodbatlas_organization",
    "project_id": "mongodbatlas_project",
    "cluster_name": "mongodbatlas_advanced_cluster",
}
SKIP_NODES: set[str] = {"mongodbatlas_cluster", "mongodbatlas_flex_cluster"}
FORCE_INTERNAL_NODES: set[str] = {"mongodbatlas_project_ip_access_list"}


def is_v2_example_dir(example_dir: Path) -> bool:
    parent_dir = example_dir.parent.name
    grand_parent_dir = example_dir.parent.parent.name
    return parent_dir in v2_parent_dir or grand_parent_dir in v2_grand_parent_dirs


def tf_dep_graph(
    repo_path: Path = REPO_PATH_ARG,
    skip_names: list[str] = SKIP_EXAMPLES_DIRS_OPTION,
):
    settings = TfDepSettings.from_env()
    output_dir = settings.static_root
    logger.info(f"Using output directory: {output_dir}")
    example_dirs = get_example_directories(repo_path, skip_names)
    logger.info(f"example_dirs: \n{'\n'.join(str(d) for d in sorted(example_dirs))}")
    with new_task("Find terraform graphs", total=len(example_dirs)) as task:
        atlas_graph = parse_graphs(example_dirs, task)
    with new_task("Dump graph"):
        graph_yaml = atlas_graph.dump_yaml()
        ensure_parents_write_text(settings.atlas_graph_path, graph_yaml)
        logger.info(f"Atlas graph dumped to {settings.atlas_graph_path}")


def print_edges(graph: pydot.Dot):
    edges = graph.get_edges()
    for edge in edges:
        logger.info(f"{edge.get_source()} -> {edge.get_destination()}")


class ResourceParts(NamedTuple):
    resource_type: str
    resource_name: str

    @property
    def provider_name(self) -> str:
        return self.resource_type.split("_")[0]


class ResourceRef(BaseModel):
    full_ref: str

    def _resource_parts(self) -> ResourceParts:
        match self.full_ref.split("."):
            case [resource_type, resource_name] if "_" in resource_type:
                return ResourceParts(resource_type, resource_name)
            case [*_, resource_type, resource_name] if "_" in resource_type:
                return ResourceParts(resource_type, resource_name)
        raise ValueError(f"Invalid resource reference: {self.full_ref}")

    @property
    def provider_name(self) -> str:
        return self._resource_parts().provider_name

    @property
    def is_external(self) -> bool:
        return self.provider_name != ATLAS_PROVIDER_NAME

    @property
    def is_atlas_resource(self) -> bool:
        return not self.is_module and not self.is_data and self.provider_name == ATLAS_PROVIDER_NAME

    @property
    def is_module(self) -> bool:
        return self.full_ref.startswith(MODULE_PREFIX)

    @property
    def is_data(self) -> bool:
        return self.full_ref.startswith(DATA_PREFIX)

    @property
    def resource_type(self) -> str:
        return self._resource_parts().resource_type


class EdgeParsed(BaseModel):
    parent: ResourceRef
    child: ResourceRef

    @classmethod
    def from_edge(cls, edge: pydot.Edge) -> "EdgeParsed":
        return cls(
            # edges shows from child --> parent, so we reverse the order
            parent=ResourceRef(full_ref=edge_plain(edge.get_destination())),
            child=ResourceRef(full_ref=edge_plain(edge.get_source())),
        )

    @property
    def has_module_edge(self) -> bool:
        return self.parent.is_module or self.child.is_module

    @property
    def has_data_edge(self) -> bool:
        return self.parent.is_data or self.child.is_data

    @property
    def is_resource_edge(self) -> bool:
        return not self.has_module_edge and not self.has_data_edge

    @property
    def is_external_to_internal_edge(self) -> bool:
        return self.parent.is_external and self.child.is_atlas_resource

    @property
    def is_internal_atlas_edge(self) -> bool:
        return self.parent.is_atlas_resource and self.child.is_atlas_resource


def edge_plain(edge_endpoint: pydot.EdgeEndpoint) -> str:
    return str(edge_endpoint).strip('"').strip()


def edge_src_dest(edge: pydot.Edge) -> tuple[str, str]:
    """Get the source and destination of the edge as plain strings."""
    return edge_plain(edge.get_source()), edge_plain(edge.get_destination())


def skip_variable_edge(src: str, dst: str) -> bool:
    # sourcery skip: assign-if-exp, boolean-if-exp-identity, reintroduce-else, remove-unnecessary-cast
    if src == dst:
        return True
    if src == "mongodbatlas_advanced_cluster" and "_cluster" in dst:
        return True
    return False


class AtlasGraph(Entity):
    # atlas_resource_type -> set[atlas_resource_type]
    parent_child_edges: dict[str, set[str]] = Field(default_factory=lambda: defaultdict(set))
    # atlas_resource_type -> set[external_resource_type]
    external_parents: dict[str, set[str]] = Field(default_factory=lambda: defaultdict(set))
    deprecated_resource_types: set[str] = Field(default_factory=set)

    def all_parents(self, child: str) -> Iterable[str]:
        for parent, children in self.parent_child_edges.items():
            if child in children:
                yield parent

    def dump_yaml(self) -> str:
        parent_child_edges = {name: sorted(children) for name, children in sorted(self.parent_child_edges.items())}
        external_parents = {name: sorted(parents) for name, parents in sorted(self.external_parents.items())}
        return dump(
            {
                "parent_child_edges": parent_child_edges,
                "external_parents": external_parents,
            },
            format="yaml",
        )

    @property
    def all_internal_nodes(self) -> set[str]:
        return set(flat_map([src] + list(dsts) for src, dsts in self.parent_child_edges.items()))

    def iterate_internal_edges(self) -> Iterable[tuple[str, str]]:
        for parent, children in self.parent_child_edges.items():
            for child in children:
                yield parent, child

    @property
    def all_external_nodes(self) -> set[str]:
        return set(flat_map([src] + list(dsts) for src, dsts in self.external_parents.items()))

    def iterate_external_edges(self) -> Iterable[tuple[str, str]]:
        for child, parents in self.external_parents.items():
            for parent in parents:
                yield parent, child

    def add_edges(self, edges: list[pydot.Edge]):
        for edge in edges:
            parsed = EdgeParsed.from_edge(edge)
            parent = parsed.parent
            child = parsed.child
            if parsed.is_internal_atlas_edge:
                self.parent_child_edges[parent.resource_type].add(child.resource_type)
                # edges shows from child --> parent, so we reverse the order
            elif parsed.is_external_to_internal_edge:
                if parent.provider_name in {"random", "cedar"}:
                    continue  # skip random provider edges
                self.external_parents[child.resource_type].add(parent.resource_type)

    def add_variable_edges(self, example_dir: Path) -> None:
        """Use the variables to find the resource dependencies."""
        if not (variables := find_variables(example_dir / "variables.tf")):
            return
        usages = find_variable_resource_type_usages(set(variables), example_dir)
        for variable, resource_types in usages.items():
            if parent_type := VARIABLE_RESOURCE_MAPPING.get(variable):
                for child_type in resource_types:
                    if skip_variable_edge(parent_type, child_type):
                        continue
                    if child_type.startswith(ATLAS_PROVIDER_NAME):
                        logger.info(f"Adding variable edge: {parent_type} -> {child_type}")
                        self.parent_child_edges[parent_type].add(child_type)


def parse_graphs(example_dirs: list[Path], task: new_task, max_workers: int = 16, max_dirs: int = 9999) -> AtlasGraph:
    atlas_graph = AtlasGraph()
    with run_pool("parse example graphs", total=len(example_dirs)) as executor:
        futures = {
            executor.submit(parse_graph, example_dir): example_dir
            for i, example_dir in enumerate(example_dirs)
            if i < max_dirs
        }
        graphs = {}
        for future in futures:
            try:
                example_dir, graph_output = future.result()
            except ShellError as e:
                logger.error(f"Error parsing graph for {futures[future]}: {e}")
                continue
            except KeyboardInterrupt:
                logger.error("KeyboardInterrupt received, stopping graph parsing.")
                stop_runs_and_pool("KeyboardInterrupt", immediate=True)
                break
            try:
                graph = graphs[example_dir] = parse_graph_output(example_dir, graph_output)
            except GraphParseError as e:
                logger.error(e)
                continue
            atlas_graph.add_edges(graph.get_edges())
            atlas_graph.add_variable_edges(example_dir)
            task.update(advance=1)
    return atlas_graph


class GraphParseError(Exception):
    def __init__(self, example_dir: Path, message: str):
        self.example_dir = example_dir
        super().__init__(f"Failed to parse graph for {example_dir}: {message}")


def parse_graph_output(example_dir: Path, graph_output: str, verbose: bool = False) -> pydot.Dot:
    assert graph_output, f"Graph output is empty for {example_dir}"
    dots = pydot.graph_from_dot_data(graph_output)  # not thread safe, so we use the main thread here instead
    if not dots:
        raise GraphParseError(example_dir, f"No graphs found in the output:\n{graph_output}")
    assert len(dots) == 1, f"Expected one graph for {example_dir}, got {len(dots)}"
    graph = dots[0]
    edges = graph.get_edges()
    if not edges:
        logger.info(f"No edges found in graph for {example_dir}")
    if verbose:
        print_edges(graph)
    return graph


class EmptyGraphOutputError(Exception):
    """Raised when the graph output is empty."""

    def __init__(self, example_dir: Path):
        self.example_dir = example_dir
        super().__init__(f"Graph output is empty for {example_dir}")


@retry(
    stop=stop_after_attempt(3),
    wait=wait_fixed(1),
    retry=retry_if_exception_type(EmptyGraphOutputError),
    reraise=True,
)
def parse_graph(example_dir: Path) -> tuple[Path, str]:
    env_vars = {
        "MONGODB_ATLAS_PREVIEW_PROVIDER_V2_ADVANCED_CLUSTER": "true" if is_v2_example_dir(example_dir) else "false",
    }
    lock_file = example_dir / ".terraform.lock.hcl"
    if not lock_file.exists():
        run_and_wait("terraform init", cwd=example_dir, env=env_vars)
    run = run_and_wait("terraform graph", cwd=example_dir, env=env_vars)
    if graph_output := run.stdout_one_line:
        return example_dir, graph_output
    raise EmptyGraphOutputError(example_dir)


def typer_main():
    app = Typer()
    app.command()(tf_dep_graph)
    configure_logging(app)
    app()


if __name__ == "__main__":
    typer_main()



import logging
import re
from collections.abc import Iterable
from pathlib import Path

from model_lib import Event

logger = logging.getLogger(__name__)


def log_optional_only(repo_path: Path):
    for resource_path in (repo_path / "internal").rglob("resource_*.go"):
        for name in schema_attributes_plugin_framework(resource_path, optional=True, computed=False):
            logger.info(f"found optional-only attr={name} in {resource_path}")


def schema_attributes_plugin_framework(path: Path, *, optional: bool = False, computed: bool = False) -> Iterable[str]:
    text = path.read_text()
    if "github.com/hashicorp/terraform-plugin-framework/resource/schema" not in text:
        return
    for attr in iterate_schema_attributes(text, "MapAttribute", "ListAttribute"):
        if attr.optional == optional and attr.computed == computed:
            yield attr.name


class SchemaAttr(Event):
    name: str
    optional: bool = False
    computed: bool = False
    attribute_type: str


def as_regex(attribute_type: str) -> re.Pattern:
    return re.compile(
        rf'"(?P<name>[\w_]+)": (?P<attribute_type>schema\.{attribute_type})\{{(?P<brackets>[^}}]+)\}}',
        re.MULTILINE,
    )


def unpack_flags(brackets: str) -> dict[str, bool]:
    return {
        "optional": "Optional:" in brackets,
        "computed": "Computed:" in brackets,
    }


def iterate_schema_attributes(text: str, *attribute_types: str) -> Iterable[SchemaAttr]:
    for attr_type in attribute_types:
        pattern = as_regex(attr_type)
        m: re.Match
        for m in pattern.finditer(text):
            match_dict = m.groupdict()
            flags = unpack_flags(match_dict.pop("brackets"))
            yield SchemaAttr(**match_dict, **flags)  # type: ignore



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import ClassVar, Self

from model_lib import Entity, dump, field_names, parse_model
from motor.motor_asyncio import AsyncIOMotorCollection
from pydantic import model_validator
from zero_3rdparty.file_utils import ensure_parents_write_text
from zero_3rdparty.iter_utils import ignore_falsy

from atlas_init.cli_tf.go_test_run import GoTestRun
from atlas_init.cli_tf.go_test_tf_error import (
    ErrorClassAuthor,
    ErrorDetailsT,
    GoTestAPIError,
    GoTestError,
    GoTestErrorClass,
    GoTestErrorClassification,
    GoTestResourceCheckError,
)
from atlas_init.crud.mongo_client import get_collection, init_mongo
from atlas_init.crud.mongo_utils import MongoQueryOperation, create_or_replace, dump_with_id
from atlas_init.repos.path import TFResoure, terraform_resources
from atlas_init.settings.env_vars import AtlasInitSettings

logger = logging.getLogger(__name__)


def crud_dir(settings: AtlasInitSettings) -> Path:
    return settings.static_root / "crud"


@dataclass
class TFResources:
    resources: list[TFResoure] = field(default_factory=list)

    def find_test_resources(self, test: GoTestRun) -> list[str]:
        found_resources = []
        for resource in self.resources:
            url = test.package_url
            if url and url.endswith(resource.package_rel_path):
                found_resources.append(resource.name)
        return found_resources


def read_tf_resources(settings: AtlasInitSettings, repo_path: Path, branch: str) -> TFResources:
    return TFResources(resources=terraform_resources(repo_path))


class TFErrors(Entity):
    errors: list[GoTestError] = field(default_factory=list)

    @model_validator(mode="after")
    def sort_errors(self) -> TFErrors:
        self.errors.sort()
        return self

    def look_for_existing_classifications(self, error: GoTestError) -> tuple[GoTestErrorClass, GoTestErrorClass] | None:
        for candidate in self.errors:
            if error.match(candidate) and (classifications := candidate.classifications):
                logger.info(f"found existing classification for {error.run.name}: {classifications}")
                return classifications

    def classified_errors(self) -> list[GoTestError]:
        return [error for error in self.errors if error.classifications is not None]


def read_tf_errors(settings: AtlasInitSettings) -> TFErrors:
    path = crud_dir(settings) / "tf_errors.yaml"
    return parse_model(path, TFErrors) if path.exists() else TFErrors()


def read_tf_errors_for_day(settings: AtlasInitSettings, branch: str, date: datetime) -> list[GoTestError]:
    raise NotImplementedError


def store_or_update_tf_errors(settings: AtlasInitSettings, errors: list[GoTestError]) -> None:
    existing = read_tf_errors(settings)
    new_error_ids = {error.run.id for error in errors}
    existing_without_new = [error for error in existing.errors if error.run.id not in new_error_ids]
    all_errors = existing_without_new + errors
    yaml_dump = dump(TFErrors(errors=all_errors), "yaml")
    ensure_parents_write_text(crud_dir(settings) / "tf_errors.yaml", yaml_dump)


def read_tf_error_by_run(settings: AtlasInitSettings, run: GoTestRun) -> GoTestError | None:
    errors = read_tf_errors(settings)
    return next((error for error in errors.errors if error.run.id == run.id), None)


class TFTestRuns(Entity):
    test_runs: list[GoTestRun] = field(default_factory=list)

    @model_validator(mode="after")
    def sort_test_runs(self) -> TFTestRuns:
        self.test_runs.sort()
        return self


def read_tf_test_runs(settings: AtlasInitSettings) -> list[GoTestRun]:
    path = crud_dir(settings) / "tf_test_runs.yaml"
    return parse_model(path, TFTestRuns).test_runs if path.exists() else []


def read_tf_tests_for_day(settings: AtlasInitSettings, branch: str, date: datetime) -> list[GoTestRun]:
    start_date = date.replace(hour=0, minute=0, second=0, microsecond=0)
    end_date = start_date.replace(hour=23, minute=59, second=59, microsecond=999999)
    return read_tf_tests(settings, branch, start_date, end_date)


def read_tf_tests(
    settings: AtlasInitSettings, branch: str, start_date: datetime, end_date: datetime | None = None
) -> list[GoTestRun]:
    raise NotImplementedError


async def init_mongo_dao(settings: AtlasInitSettings) -> MongoDao:
    dao = MongoDao(settings=settings)
    return await dao.connect()


class GoTestRunNotFound(Exception):
    def __init__(self, run_id: str) -> None:
        self.run_id = run_id
        super().__init__(run_id)


@dataclass
class MongoDao:
    settings: AtlasInitSettings
    property_keys_run: ClassVar[list[str]] = ["group_name"]

    @cached_property
    def runs(self) -> AsyncIOMotorCollection:
        return get_collection(GoTestRun)

    @cached_property
    def classifications(self) -> AsyncIOMotorCollection:
        return get_collection(GoTestErrorClassification)

    @cached_property
    def _field_names_runs(self) -> set[str]:
        return set(field_names(GoTestRun)) | set(self.property_keys_run)

    async def connect(self) -> Self:
        await init_mongo(
            mongo_url=self.settings.mongo_url,
            db_name=self.settings.mongo_database,
        )
        return self

    async def store_tf_test_runs(self, test_runs: list[GoTestRun]) -> list[GoTestRun]:
        if not test_runs:
            return []
        col = self.runs
        tasks = []
        loop = asyncio.get_event_loop()
        for run in test_runs:
            dumped = dump_with_id(run, id=run.id, dt_keys=["ts", "finish_ts"], property_keys=self.property_keys_run)
            tasks.append(loop.create_task(create_or_replace(col, dumped)))
        await asyncio.gather(*tasks)
        return test_runs

    async def read_tf_tests_for_day(self, branch: str, date: datetime) -> list[GoTestRun]:
        start_date = date.replace(hour=0, minute=0, second=0, microsecond=0)
        end_date = start_date.replace(hour=23, minute=59, second=59, microsecond=999999)
        query = {
            "branch": branch,
            "ts": {MongoQueryOperation.gte: start_date, MongoQueryOperation.lte: end_date},
        }
        return await self._find_runs(query)

    async def _find_runs(self, query: dict) -> list[GoTestRun]:
        runs = []
        async for raw_run in self.runs.find(query):
            runs.append(self._parse_run(raw_run))
        return runs

    async def read_error_classifications(
        self, run_ids: list[str] | None = None
    ) -> dict[str, GoTestErrorClassification]:
        run_ids = run_ids or []
        if not run_ids:
            return {}
        query = {"_id": {MongoQueryOperation.in_: run_ids}}
        return await self._find_classifications(query)

    async def _find_classifications(self, query: dict) -> dict[str, GoTestErrorClassification]:
        classifications: dict[str, GoTestErrorClassification] = {}
        async for raw_error in self.classifications.find(query):
            run_id = raw_error.pop("_id", None)
            classification = parse_model(raw_error, t=GoTestErrorClassification)
            classifications[run_id] = classification
        return classifications

    async def read_similar_error_classifications(
        self, details: ErrorDetailsT, *, author_filter: ErrorClassAuthor | None = None
    ) -> dict[str, GoTestErrorClassification]:
        query = {}
        if author_filter:
            query["author"] = {MongoQueryOperation.eq: author_filter}
        match details:
            case GoTestAPIError(
                api_error_code_str=api_error_code_str,
                api_method=api_method,
                api_response_code=api_response_code,
                api_path_normalized=api_path_normalized,
            ) if api_path_normalized:
                query |= {
                    "details.api_error_code_str": {MongoQueryOperation.eq: api_error_code_str},
                    "details.api_method": {MongoQueryOperation.eq: api_method},
                    "details.api_response_code": {MongoQueryOperation.eq: api_response_code},
                    "details.api_path_normalized": {MongoQueryOperation.eq: api_path_normalized},
                }
            case GoTestResourceCheckError(
                tf_resource_name=tf_resource_name,
                tf_resource_type=tf_resource_type,
                step_nr=step_nr,
                check_errors=check_errors,
                test_name=test_name,
            ):
                query |= {
                    "details.tf_resource_name": {MongoQueryOperation.eq: tf_resource_name},
                    "details.tf_resource_type": {MongoQueryOperation.eq: tf_resource_type},
                    "details.step_nr": {MongoQueryOperation.eq: step_nr},
                    "test_name": {MongoQueryOperation.eq: test_name},
                }
                classifications = await self._find_classifications(query)
                return {
                    run_id: classification
                    for run_id, classification in classifications.items()
                    if isinstance(classification.details, GoTestResourceCheckError)
                    and classification.details.check_errors_match(check_errors)
                }
            case _:
                return {}  # todo: vector search to match on error output
        return await self._find_classifications(query)

    async def add_classification(self, classification: GoTestErrorClassification) -> bool:
        """Returns is_new"""
        raw = dump_with_id(classification, id=classification.run_id, dt_keys=["ts"])
        return await create_or_replace(self.classifications, raw)

    async def read_tf_test_run(self, run_id: str) -> GoTestRun:
        raw = await self.runs.find_one({"_id": run_id})
        if raw is None:
            raise GoTestRunNotFound(run_id)
        return self._parse_run(raw)

    def _parse_run(self, raw: dict) -> GoTestRun:
        raw.pop("_id")
        for key in self.property_keys_run:
            raw.pop(key, None)  # Remove properties that are not part of the model
        return parse_model(raw, t=GoTestRun)

    async def read_run_history(
        self,
        test_name: str,
        branches: list[str] | None = None,
        package_url: str | None = None,
        group_name: str | None = None,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        envs: list[str] | None = None,
    ) -> list[GoTestRun]:
        eq = MongoQueryOperation.eq
        query = {
            "name": {eq: test_name},
        }
        eq_parts = {
            "package_url": {eq: package_url} if package_url else None,
            "group_name": {eq: group_name} if group_name else None,
        }
        in_op = MongoQueryOperation.in_
        in_parts = {
            "branch": {in_op: branches} if branches else None,
            "env": {in_op: envs} if envs else None,
        }
        date_parts = {
            "ts": ignore_falsy(
                **{
                    MongoQueryOperation.lte: end_date or None,
                    MongoQueryOperation.gte: start_date or None,
                }
            )
        }
        query |= ignore_falsy(**eq_parts, **in_parts, **date_parts)
        if invalid_fields := set(query) - self._field_names_runs:
            raise ValueError(f"Invalid fields in query: {invalid_fields}")
        return await self._find_runs(query)



import logging
import os
import re
from pathlib import Path

import typer
from model_lib import Entity
from pydantic import Field
from zero_3rdparty.file_utils import ensure_parents_write_text

from atlas_init.cli_cfn.files import create_sample_file_from_input
from atlas_init.cli_helper.run import (
    run_binary_command_is_ok,
)
from atlas_init.cli_helper.run_manager import RunManager
from atlas_init.cli_root import is_dry_run
from atlas_init.repos.path import Repo, ResourcePaths, find_paths
from atlas_init.settings.env_vars import AtlasInitSettings, init_settings
from atlas_init.settings.env_vars_generated import AWSSettings

logger = logging.getLogger(__name__)


class RunContractTest(Entity):
    resource_path: Path
    repo_path: Path
    cfn_region: str
    aws_profile: str
    skip_build: bool = False
    dry_run: bool = Field(default_factory=is_dry_run)
    only_names: list[str] | None = None

    @property
    def run_tests_command(self) -> tuple[str, str]:
        if self.only_names:
            names = " ".join(f"-k {name}" for name in self.only_names)
            return (
                "cfn",
                f"test --function-name TestEntrypoint --verbose --region {self.cfn_region} -- {names}",
            )
        return (
            "cfn",
            f"test --function-name TestEntrypoint --verbose --region {self.cfn_region}",
        )


class RunContractTestOutput(Entity):
    sam_local_logs: str
    sam_local_exit_code: int
    contract_test_ok: bool
    rpdk_log: str


class CreateContractTestInputs(Entity):
    resource_path: Path
    log_group_name: str


class CFNBuild(Entity):
    resource_path: Path
    dry_run: bool = Field(default_factory=is_dry_run)
    is_debug: bool = False
    tags: str = "logging callback metrics scheduler"
    cgo: int = 0
    goarch: str = "amd64"
    goos: str = "linux"
    git_sha: str = "local"
    ldflags: str = "-s -w -X github.com/mongodb/mongodbatlas-cloudformation-resources/util.defaultLogLevel=info -X github.com/mongodb/mongodbatlas-cloudformation-resources/version.Version=${CFNREP_GIT_SHA}"

    @property
    def extra_env(self) -> dict[str, str]:
        return {"GOOS": self.goos, "CGO_ENABLED": str(self.cgo), "GOARCH": self.goarch}

    @property
    def flags(self) -> str:
        return self.ldflags.replace("${CFNREP_GIT_SHA}", self.git_sha)

    @property
    def command_build(self) -> str:
        return f'build -ldflags="{self.flags}" -tags="{self.tags}" -o bin/bootstrap cmd/main.go'

    @property
    def cfn_generate(self) -> str:
        return "generate"

    @property
    def commands(self) -> list[tuple[str, str]]:
        return [
            ("cfn", self.cfn_generate),
            ("go", self.command_build),
        ]


def contract_test_cmd(
    only_names: list[str] = typer.Option(None, "-n", "--only-names", help="only run these contract tests"),
):
    result = contract_test(only_names=only_names)
    if result.contract_test_ok:
        logger.info("contract tests passed 🥳")
    else:
        logger.error("contract tests failed 💥")
        logger.error(
            f"function logs (exit_code={result.sam_local_exit_code}):\n {result.sam_local_logs}\n\nRPDK logs:\n{result.rpdk_log[-10_000:]}"
        )
        raise typer.Exit(1)
    return result


def contract_test(
    settings: AtlasInitSettings | None = None,
    resource_paths: ResourcePaths | None = None,
    only_names: list[str] | None = None,
):
    settings = settings or init_settings(AWSSettings)
    resource_paths = resource_paths or find_paths(Repo.CFN)
    resource_name = resource_paths.resource_name
    create_inputs = CreateContractTestInputs(
        resource_path=resource_paths.resource_path,
        log_group_name=f"mongodb-atlas-{resource_name}-logs",
    )
    create_response = create_contract_test_inputs(create_inputs)
    create_response.log_input_files(logger)
    aws_settings = AWSSettings.from_env()
    run_contract_test = RunContractTest(
        resource_path=resource_paths.resource_path,
        repo_path=resource_paths.repo_path,
        aws_profile=aws_settings.AWS_PROFILE,
        cfn_region=settings.cfn_region(aws_settings.AWS_REGION),
        only_names=only_names,
    )
    if run_contract_test.skip_build:
        logger.info("skipping build")
    else:
        build_event = CFNBuild(resource_path=resource_paths.resource_path)
        build(build_event)
        logger.info("build ok ✅")
    return run_contract_tests(run_contract_test)


class CreateContractTestInputsResponse(Entity):
    input_files: list[Path]
    sample_files: list[Path]

    def log_input_files(self, logger: logging.Logger):
        inputs = self.input_files
        if not inputs:
            logger.warning("no input files created")
            return
        inputs_dir = self.input_files[0].parent
        logger.info(f"{len(inputs)} inputs created in '{inputs_dir}'")
        logger.info("\n".join(f"'{file.name}'" for file in self.input_files))


def create_contract_test_inputs(
    event: CreateContractTestInputs,
) -> CreateContractTestInputsResponse:
    inputs_dir = event.resource_path / "inputs"
    samples_dir = event.resource_path / "samples"
    test_dir = event.resource_path / "test"
    sample_files = []
    input_files = []
    for template in sorted(test_dir.glob("*.template.json")):
        template_file = template.read_text()
        template_file = file_replacements(template_file, template.name)
        inputs_file = inputs_dir / template.name.replace(".template", "")
        ensure_parents_write_text(inputs_file, template_file)
        input_files.append(inputs_file)
        sample_file = create_sample_file_from_input(samples_dir, event.log_group_name, inputs_file)
        sample_files.append(sample_file)
    return CreateContractTestInputsResponse(input_files=input_files, sample_files=sample_files)


def file_replacements(text: str, file_name: str) -> str:
    for match in re.finditer(r"\${(\w+)}", text):
        var_name = match.group(1)
        if env_value := os.environ.get(var_name):
            text = text.replace(match.group(0), env_value)
        else:
            logger.warning(f"found placeholder {match.group(0)} in {file_name} but no replacement")
    return text


def build(event: CFNBuild):
    for binary, command in event.commands:
        is_ok = run_binary_command_is_ok(
            binary,
            command,
            cwd=event.resource_path,
            logger=logger,
            dry_run=event.dry_run,
            env={**os.environ, **event.extra_env},
        )
        if not is_ok:
            logger.critical(f"failed to run {binary} {command}")
            raise typer.Exit(1)


def run_contract_tests(event: RunContractTest) -> RunContractTestOutput:
    with RunManager(dry_run=event.dry_run) as manager:
        manager.set_timeouts(3)
        resource_path = event.resource_path
        run_future = manager.run_process_wait_on_log(
            f"local start-lambda --skip-pull-image --region {event.cfn_region}",
            binary="sam",
            cwd=resource_path,
            logger=logger,
            line_in_log="Running on http://",
            timeout=60,
        )
        binary, test_cmd = event.run_tests_command
        test_result_ok = run_binary_command_is_ok(
            binary,
            test_cmd,
            cwd=resource_path,
            logger=logger,
            dry_run=event.dry_run,
        )
        extra_log = resource_path / "rpdk.log"
        log_content = extra_log.read_text() if extra_log.exists() else ""
    sam_local_result = run_future.result(timeout=1)
    return RunContractTestOutput(
        sam_local_logs=sam_local_result.result_str,
        sam_local_exit_code=sam_local_result.exit_code or -1,
        contract_test_ok=test_result_ok,
        rpdk_log=log_content,
    )



import logging
import os
import sys
from functools import partial

import typer

from atlas_init import running_in_repo
from atlas_init.cli_cfn.app import app as app_cfn
from atlas_init.cli_helper.run import add_to_clipboard
from atlas_init.cli_root import set_dry_run
from atlas_init.cli_tf.app import app as app_tf
from atlas_init.cloud.aws import download_from_s3, upload_to_s3
from atlas_init.settings.env_vars import (
    DEFAULT_PROFILE,
    ENV_CLIPBOARD_COPY,
    ENV_PROFILE,
    ENV_PROJECT_NAME,
    ENV_S3_PROFILE_BUCKET,
    init_settings,
)

logger = logging.getLogger(__name__)


def sync_on_done(return_value, s3_profile_bucket: str = "", use_clipboard: str = "", **kwargs):
    logger.debug(f"sync_on_done return_value={return_value} and {kwargs}")
    settings = init_settings(skip_ambiguous_check=True)
    if s3_profile_bucket:
        logger.info(f"using s3 bucket for profile sync: {s3_profile_bucket}")
        upload_to_s3(settings.profile_dir, s3_profile_bucket)
    if use_clipboard:
        settings = settings or init_settings()
        match use_clipboard:
            case "manual":
                env_path = settings.env_file_manual
            case _:
                env_path = settings.env_vars_generated
        if env_path.exists():
            clipboard_content = "\n".join(f"export {line}" for line in env_path.read_text().splitlines())
            add_to_clipboard(clipboard_content, logger)
            logger.info(f"loaded env-vars from {env_path} to clipboard ✅")


app = typer.Typer(
    name="atlas_init",
    invoke_without_command=True,
    no_args_is_help=True,
    result_callback=sync_on_done,
)
app.add_typer(app_cfn, name="cfn")
app.add_typer(app_tf, name="tf")

app_command = partial(
    app.command,
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)


def extra_root_commands():
    from atlas_init.cli_root import go_test, trigger, mms_released

    assert trigger
    assert go_test
    assert mms_released


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    log_level: str = typer.Option("INFO", help="use one of [INFO, WARNING, ERROR, CRITICAL]"),
    profile: str = typer.Option(
        DEFAULT_PROFILE,
        "-p",
        "--profile",
        envvar=ENV_PROFILE,
        help="used to load .env_manual, store terraform state and variables, and dump .env files.",
    ),
    project_name: str = typer.Option(
        "",
        "--project",
        envvar=ENV_PROJECT_NAME,
        help="atlas project name to create",
    ),
    show_secrets: bool = typer.Option(False, help="show secrets in the logs"),
    dry_run: bool = typer.Option(False, help="dry-run mode"),
    s3_profile_bucket: str = typer.Option(
        "",
        "-s3",
        "--s3-profile-bucket",
        help="s3 bucket to store profiles will be synced before and after the command",
        envvar=ENV_S3_PROFILE_BUCKET,
    ),
    use_clipboard: str = typer.Option(
        "",
        help="add env-vars to clipboard, manual=.env-manual file, other value=.env-generated file",
        envvar=ENV_CLIPBOARD_COPY,
    ),
):
    set_dry_run(dry_run)
    if profile != DEFAULT_PROFILE:
        os.environ[ENV_PROFILE] = profile
    if project_name != "":
        os.environ[ENV_PROJECT_NAME] = project_name
    if use_clipboard:
        os.environ[ENV_CLIPBOARD_COPY] = use_clipboard
    is_running_in_repo = running_in_repo()
    logger.info(f"running in atlas-init repo: {is_running_in_repo} python location:{sys.executable}")
    logger.info(f"in the app callback, log-level: {log_level}, command: {format_cmd(ctx)}")
    if s3_bucket := s3_profile_bucket:
        logger.info(f"using s3 bucket for profile sync: {s3_bucket}")
        settings = init_settings()
        download_from_s3(settings.profile_dir, s3_bucket)
    settings = init_settings()


def format_cmd(ctx: typer.Context) -> str:
    return f"'{ctx.info_name} {ctx.invoked_subcommand}'"



from __future__ import annotations

import logging
from pathlib import Path
from typing import ClassVar, NamedTuple

from ask_shell import new_task, run_and_wait
from model_lib import IgnoreFalsy, dump
from pydantic import Field, RootModel
from zero_3rdparty.file_utils import ensure_parents_write_text
from zero_3rdparty.str_utils import instance_repr

from atlas_init.tf_ext.args import REPO_PATH_ARG, SKIP_EXAMPLES_DIRS_OPTION
from atlas_init.tf_ext.constants import ATLAS_PROVIDER_NAME
from atlas_init.tf_ext.paths import (
    ResourceTypes,
    find_resource_types_with_usages,
    find_variables,
    get_example_directories,
    is_variable_name_external,
)
from atlas_init.tf_ext.settings import TfDepSettings

logger = logging.getLogger(__name__)


def tf_vars(
    repo_path: Path = REPO_PATH_ARG,
    skip_names: list[str] = SKIP_EXAMPLES_DIRS_OPTION,
):
    settings = TfDepSettings.from_env()
    logger.info(f"Analyzing Terraform variables in repository: {repo_path}")
    example_dirs = get_example_directories(repo_path, skip_names)
    assert example_dirs, "No example directories found. Please check the repository path and skip names."
    with new_task("Parsing provider schema") as task:
        resource_types, resource_types_deprecated = parse_schema_resource_types(example_dirs[0])
        ensure_parents_write_text(settings.schema_resource_types_path, dump(sorted(resource_types), format="yaml"))
        logger.info(f"Provider schema resource types written to {settings.schema_resource_types_path}")
        ensure_parents_write_text(
            settings.schema_resource_types_deprecated_path, dump(sorted(resource_types_deprecated), format="yaml")
        )
        logger.info(
            f"Provider schema deprecated resource types written to {settings.schema_resource_types_deprecated_path}"
        )
    logger.info(f"Found {len(resource_types)} resource types in the provider schema.: {', '.join(resource_types)}")
    with new_task("Parsing variables from examples") as task:
        update_variables(settings, example_dirs, task)
    with new_task("Parsing resource types from examples", total=len(example_dirs)) as task:
        example_resource_types = update_resource_types(settings, example_dirs, task)
    if missing_example_resource_types := set(resource_types) - set(example_resource_types.root):
        logger.warning(f"Missing resource types in examples:\n{'\n'.join(sorted(missing_example_resource_types))}")


def parse_provider_resurce_schema(schema: dict, provider_name: str) -> dict:
    schemas = schema.get("provider_schemas", {})
    for provider_url, provider_schema in schemas.items():
        if provider_url.endswith(provider_name):
            return provider_schema.get("resource_schemas", {})
    raise ValueError(f"Provider '{provider_name}' not found in schema.")


class TfVarUsage(IgnoreFalsy):
    name: str = Field(..., description="Name of the Terraform variable.")
    descriptions: set[str] = Field(default_factory=set, description="Set of descriptions for the variable.")
    example_paths: list[Path] = Field(
        default_factory=list, description="List of example files where the variable is used."
    )

    PARENT_DIR: ClassVar[str] = "examples"

    def update(self, variable_description: str | None, example_dir: Path):
        if variable_description and variable_description not in self.descriptions:
            self.descriptions.add(variable_description)
        assert f"/{self.PARENT_DIR}/" in str(example_dir), "Example directory must be under 'examples/'"
        if example_dir not in self.example_paths:
            self.example_paths.append(example_dir)

    @property
    def paths_str(self) -> str:
        return ", ".join(str(path).split(self.PARENT_DIR)[1] for path in self.example_paths)

    def __str__(self):
        return instance_repr(self, ["name", "descriptions", "paths_str"])

    def dump_dict_modifier(self, payload: dict) -> dict:
        payload["descriptions"] = sorted(self.descriptions)
        payload["example_paths"] = sorted(self.example_paths)
        return payload


class TfVarsUsage(RootModel[dict[str, TfVarUsage]]):
    def add_variable(self, variable: str, variable_description: str | None, example_dir: Path):
        if variable not in self.root:
            self.root[variable] = TfVarUsage(name=variable, example_paths=[])
        self.root[variable].update(variable_description, example_dir)

    def external_vars(self) -> TfVarsUsage:
        return type(self)(root={name: usage for name, usage in self.root.items() if is_variable_name_external(name)})


def vars_usage_dumping(variables: TfVarsUsage) -> str:
    vars_model = variables.model_dump()
    vars_model = dict(sorted(vars_model.items()))
    return dump(vars_model, format="yaml")


def update_resource_types(settings: TfDepSettings, example_dirs: list[Path], task: new_task) -> ResourceTypes:
    resource_types = ResourceTypes(root={})
    for example_dir in example_dirs:
        example_resources = find_resource_types_with_usages(example_dir)
        for resource_type, usages in example_resources.root.items():
            resource_types.add_resource_type(resource_type, usages.example_files, usages.variable_usage)
        task.update(advance=1)
    logger.info(f"Found {len(resource_types.root)} resource types in the examples.")
    resource_types_yaml = resource_types_dumping(resource_types)
    ensure_parents_write_text(settings.resource_types_file_path, resource_types_yaml)
    logger.info(f"Resource types usage written to {settings.resource_types_file_path}")
    atlas_external_resource_types = resource_types.atlas_resource_type_with_external_var_usages()
    logger.info(f"Found {len(atlas_external_resource_types.root)} Atlas resource types with external variable usages.")
    atlas_external_resource_types_yaml = resource_types_dumping(atlas_external_resource_types, with_external=True)
    ensure_parents_write_text(settings.resource_types_external_file_path, atlas_external_resource_types_yaml)
    logger.info(
        f"Atlas resource types with external variable usages written to {settings.resource_types_external_file_path}"
    )
    return resource_types


def resource_types_dumping(resource_types: ResourceTypes, with_external: bool = False) -> str:
    resource_types_model = resource_types.dump_with_external_vars() if with_external else resource_types.model_dump()
    return dump(dict(sorted(resource_types_model.items())), format="yaml")


def update_variables(settings: TfDepSettings, example_dirs: list[Path], task: new_task):
    variables = parse_all_variables(example_dirs, task)
    logger.info(f"Found {len(variables.root)} variables in the examples.")
    vars_yaml = vars_usage_dumping(variables)
    ensure_parents_write_text(settings.vars_file_path, vars_yaml)
    logger.info(f"Variables usage written to {settings.vars_file_path}")
    external_vars = variables.external_vars()
    if external_vars.root:
        logger.info(f"Found {len(external_vars.root)} external variables: {', '.join(external_vars.root.keys())}")
        external_vars_yaml = vars_usage_dumping(external_vars)
        ensure_parents_write_text(settings.vars_external_file_path, external_vars_yaml)
        logger.info(f"External variables usage written to {settings.vars_external_file_path}")


class ResourceTypesSchema(NamedTuple):
    resource_types: list[str]
    deprecated_resource_types: list[str]


def parse_schema_resource_types(example_dir: Path) -> ResourceTypesSchema:
    schema_run = run_and_wait("terraform providers schema -json", cwd=example_dir, ansi_content=False)
    parsed = schema_run.parse_output(dict, output_format="json")
    resource_schema = parse_provider_resurce_schema(parsed, ATLAS_PROVIDER_NAME)

    def is_deprecated(resource_details: dict) -> bool:
        return resource_details["block"].get("deprecated", False)

    deprecated_resource_types = [name for name, details in resource_schema.items() if is_deprecated(details)]
    return ResourceTypesSchema(sorted(resource_schema.keys()), sorted(deprecated_resource_types))


def parse_all_variables(examples_dirs: list[Path], task: new_task) -> TfVarsUsage:
    variables_usage = TfVarsUsage(root={})
    for example_dir in examples_dirs:
        variables_tf = example_dir / "variables.tf"
        if not variables_tf.exists():
            continue
        for variable, variable_desc in find_variables(variables_tf).items():
            variables_usage.add_variable(variable, variable_desc, example_dir)
        task.update(advance=1)
    return variables_usage



from __future__ import annotations

import logging
from collections import defaultdict
from pathlib import Path
from typing import Self

from model_lib import Entity
from pydantic import Field, RootModel
from zero_3rdparty.file_utils import iter_paths

from atlas_init.cli_tf.hcl.modifier2 import resource_types_vars_usage, safe_parse, variable_reader, variable_usages
from atlas_init.tf_ext.constants import ATLAS_PROVIDER_NAME, DEFAULT_EXTERNAL_SUBSTRINGS, DEFAULT_INTERNAL_SUBSTRINGS

logger = logging.getLogger(__name__)


def find_example_dirs(repo_path: Path) -> list[Path]:
    example_dirs: set[Path] = {
        tf_file.parent for tf_file in (repo_path / "examples").rglob("*.tf") if ".terraform" not in tf_file.parts
    }
    return sorted(example_dirs)


def get_example_directories(repo_path: Path, skip_names: list[str]):
    example_dirs = find_example_dirs(repo_path)
    logger.info(f"Found {len(example_dirs)} exaple directories in {repo_path}")
    if skip_names:
        len_before = len(example_dirs)
        example_dirs = [d for d in example_dirs if d.name not in skip_names]
        logger.info(f"Skipped {len_before - len(example_dirs)} example directories with names: {skip_names}")
    return example_dirs


def find_variables(variables_tf: Path) -> dict[str, str | None]:
    if not variables_tf.exists():
        return {}
    tree = safe_parse(variables_tf)
    if not tree:
        logger.warning(f"Failed to parse {variables_tf}")
        return {}
    return variable_reader(tree)


def find_variable_resource_type_usages(variables: set[str], example_dir: Path) -> dict[str, set[str]]:
    usages = defaultdict(set)
    for path in example_dir.glob("*.tf"):
        tree = safe_parse(path)
        if not tree:
            logger.warning(f"Failed to parse {path}")
            continue
        path_usages = variable_usages(variables, tree)
        for variable, resources in path_usages.items():
            usages[variable].update(resources)
    return usages


class ResourceVarUsage(Entity):
    var_name: str
    attribute_path: str


def is_variable_name_external(
    name: str, external_substrings: list[str] | None = None, internal_substrings: list[str] | None = None
) -> bool:
    external_substrings = external_substrings or DEFAULT_EXTERNAL_SUBSTRINGS
    internal_substrings = internal_substrings or DEFAULT_INTERNAL_SUBSTRINGS
    if any(substring in name for substring in internal_substrings):
        return False
    return any(substring in name for substring in external_substrings)


class ResourceTypeUsage(Entity):
    name: str
    example_files: list[Path] = Field(default_factory=list)
    variable_usage: list[ResourceVarUsage] = Field(default_factory=list)

    def add_usage(self, example_files: list[Path], variable_usages: list[ResourceVarUsage]):
        for example_file in example_files:
            if example_file not in self.example_files:
                self.example_files.append(example_file)
        self.variable_usage.extend(variable_usages)

    @property
    def external_var_usages(self) -> list[str]:
        return [usage.var_name for usage in self.variable_usage if is_variable_name_external(usage.var_name)]


class ResourceTypes(RootModel[dict[str, ResourceTypeUsage]]):
    def add_resource_type(self, resource_type: str, example_files: list[Path], variable_usages: list[ResourceVarUsage]):
        if resource_type not in self.root:
            self.root[resource_type] = ResourceTypeUsage(name=resource_type)
        resource_type_usage = self.root[resource_type]
        resource_type_usage.add_usage(example_files, variable_usages)

    def atlas_resource_type_with_external_var_usages(self) -> Self:
        return type(self)(
            root={
                name: usage
                for name, usage in self.root.items()
                if name.startswith(ATLAS_PROVIDER_NAME) and usage.external_var_usages
            }
        )

    def dump_with_external_vars(self) -> dict[str, dict]:
        return {
            name: usages.model_dump() | {"external_var_usages": usages.external_var_usages}
            for name, usages in self.root.items()
        }


def find_resource_types_with_usages(example_dir: Path):
    output = ResourceTypes(root={})
    for path in iter_paths(example_dir, "*.tf", exclude_folder_names=[".terraform"]):
        tree = safe_parse(path)
        if not tree:
            logger.warning(f"Failed to parse {path}")
            continue
        type_var_usages = resource_types_vars_usage(tree)
        for resource_type, var_usages in type_var_usages.items():
            variable_usages = [
                ResourceVarUsage(var_name=variable_name, attribute_path=attribute_path)
                for variable_name, attribute_path in var_usages.items()
            ]
            output.add_resource_type(resource_type, example_files=[path], variable_usages=variable_usages)
    return output



import json
import logging
import time
from collections.abc import Callable
from io import StringIO
from pathlib import Path
from typing import Self

import typer
import yaml
from model_lib import Entity
from pydantic import Field, model_validator
from zero_3rdparty import file_utils

from atlas_init.cli_args import option_sdk_repo_path
from atlas_init.cli_tf.debug_logs import (
    SDKRoundtrip,
    parse_http_requests,
    parse_test_name,
)
from atlas_init.cli_tf.debug_logs_test_data import (
    RTModifier,
    create_mock_data,
    default_is_diff,
)
from atlas_init.cli_tf.debug_logs_test_data_package_config import (
    package_modifiers,
    package_must_substrings,
    package_skip_suffixes,
)
from atlas_init.repos.go_sdk import (
    api_spec_path_transformed,
    download_admin_api,
    parse_api_spec_paths,
)
from atlas_init.settings.env_vars import init_settings

logger = logging.getLogger(__name__)


class MockTFLog(Entity):
    log_path: Path
    output_dir: Path
    admin_api_path: Path
    diff_skip_suffixes: list[str] = Field(default_factory=list)
    diff_must_substrings: list[str] = Field(default_factory=list)
    keep_duplicates: bool = False
    modifiers: list[RTModifier] = Field(default_factory=list)
    package_name: str = ""
    log_diff_roundtrips: bool = False
    skip_default_package_config: bool = False

    @model_validator(mode="after")
    def ensure_paths_exist(self) -> Self:
        if not self.log_path.exists():
            raise ValueError(f"log_path: '{self.log_path}' doesn't exist")
        if not self.admin_api_path.exists():
            raise ValueError(f"admin_api_path: '{self.admin_api_path}' doesn't exist")
        if not self.output_dir.exists():
            raise ValueError(f"output_dir: '{self.output_dir}' doesn't exist")
        assert self.output_dir.name == "testdata", "output_path should be a directory named testdata"
        if (package_name := self.package_name) and not self.skip_default_package_config:
            self.modifiers.extend(package_modifiers(package_name))
            self.diff_skip_suffixes.extend(package_skip_suffixes(package_name))
            self.diff_must_substrings.extend(package_must_substrings(package_name))
        return self

    def differ(self, rt: SDKRoundtrip) -> bool:
        is_diff = default_is_diff(rt) and not any(
            rt.request.path.endswith(suffix) for suffix in self.diff_skip_suffixes
        )
        if is_diff and self.diff_must_substrings:
            return is_diff and all(substring in rt.request.path for substring in self.diff_must_substrings)
        return is_diff


def mock_tf_log(req: MockTFLog) -> Path:
    log_file_text = req.log_path.read_text()
    test_name = parse_test_name(log_file_text)
    roundtrips = parse_http_requests(log_file_text)
    logger.info(f"Found #{len(roundtrips)} roundtrips")
    if req.log_diff_roundtrips:
        log_diff_roundtrips(roundtrips, req.differ)
    api_spec_paths = parse_api_spec_paths(req.admin_api_path)
    data = create_mock_data(
        roundtrips,
        api_spec_paths,
        is_diff=req.differ,
        prune_duplicates=not req.keep_duplicates,
        modifiers=req.modifiers,
    )
    # avoid anchors
    data_json = data.model_dump_json(exclude_none=True)
    data_parsed = json.loads(data_json)
    s = StringIO()
    yaml.safe_dump(
        data_parsed,
        s,
        default_flow_style=False,
        width=100_000,
        allow_unicode=True,
        sort_keys=False,
    )
    data_yaml = s.getvalue()
    test_name = test_name.replace("TestAcc", "TestMock")
    output_path = req.output_dir / f"{test_name}.yaml"
    logger.info(f"Variables found {data.variables}")
    logger.info(f"Writing to {output_path}")
    file_utils.ensure_parents_write_text(output_path, data_yaml)
    return output_path


def mock_tf_log_cmd(
    log_path: str = typer.Argument(..., help="the path to the log file generated with TF_LOG_PATH"),
    output_testdir: str = typer.Option(
        "",
        "-o",
        "--output-testdir",
        help="the path to the output test directory, for example: internal/service/advancedclustertpf/testdata/, uses $(cwd)/testdata by default",
    ),
    sdk_repo_path_str: str = option_sdk_repo_path,
    sdk_branch: str = typer.Option("main", "-b", "--branch", help="the branch for downloading openapi spec"),
    admin_api_path: str = typer.Option(
        "", "-a", "--admin-api-path", help="the path to store/download the openapi spec"
    ),
    diff_skip_suffixes: list[str] = typer.Option(..., "-s", "--skip-suffixes", default_factory=list),
    keep_duplicates: bool = typer.Option(False, "-keep", "--keep-duplicates", help="keep duplicate requests"),
    log_diff_roundtrips: bool = typer.Option(
        False, "-l", "--log-diff-roundtrips", help="print out the roundtrips used in diffs"
    ),
    package_name: str = typer.Option("-p", "--package-name", prompt=True, help="the package name to use for modifiers"),
):
    cwd = Path.cwd()
    default_testdir = cwd / "testdata"
    resolved_admin_api_path = resolve_admin_api_path(sdk_repo_path_str, sdk_branch, admin_api_path)
    event_in = MockTFLog(
        log_path=Path(log_path),
        output_dir=Path(output_testdir) if output_testdir else default_testdir,
        admin_api_path=resolved_admin_api_path,
        diff_skip_suffixes=diff_skip_suffixes,
        keep_duplicates=keep_duplicates,
        log_diff_roundtrips=log_diff_roundtrips,
        package_name=package_name,
    )
    mock_tf_log(event_in)


def is_cache_up_to_date(cache_path: Path, cache_ttl: int) -> bool:
    if cache_path.exists():
        modified_ts = file_utils.file_modified_time(cache_path)
        if modified_ts > time.time() - cache_ttl:
            logger.info(f"using cached admin api: {cache_path} downloaded {time.time() - modified_ts:.0f}s ago")
            return True
    return False


def resolve_admin_api_path(sdk_repo_path_str: str = "", sdk_branch: str = "main", admin_api_path: str = "") -> Path:
    if admin_api_path:
        resolved_admin_api_path = Path(admin_api_path)
        if not resolved_admin_api_path.exists():
            download_admin_api(resolved_admin_api_path, sdk_branch)
    elif sdk_repo_path_str:
        sdk_repo_path = Path(sdk_repo_path_str)
        assert sdk_repo_path.exists(), f"not found sdk_repo_path={sdk_repo_path}"
        resolved_admin_api_path = api_spec_path_transformed(sdk_repo_path)
    else:
        settings = init_settings()
        resolved_admin_api_path = settings.atlas_atlas_api_transformed_yaml
        if not is_cache_up_to_date(resolved_admin_api_path, 3600):
            download_admin_api(resolved_admin_api_path, sdk_branch)
    assert resolved_admin_api_path.exists(), f"unable to resolve admin_api_path={resolved_admin_api_path}"
    assert resolved_admin_api_path.is_file(), f"not a file admin_api_path={resolved_admin_api_path}"
    return resolved_admin_api_path


def log_diff_roundtrips(roundtrips: list[SDKRoundtrip], differ: Callable[[SDKRoundtrip], bool] | None = None):
    differ = differ or default_is_diff
    diff_count = 0
    step_nr = 0
    for rt in roundtrips:
        if not differ(rt):
            continue
        if rt.step_number != step_nr:
            logger.info(f"{'-' * 80}\nStep {rt.step_number}")
            step_nr = rt.step_number
        diff_count += 1
        logger.info(
            f"\n{rt.request.method} {rt.request.path} {rt.version}\n{rt.request.text}\n{rt.response.status}-{rt.response.status_text}\n{rt.response.text}"
        )
    logger.info(f"Diffable requests: {diff_count}")



import logging
from functools import singledispatch
from queue import Queue

from atlas_init.cli_tf.schema_v2 import (
    extend_import_urls,
    go_fmt,
    import_lines,
    package_name,
)
from atlas_init.cli_tf.schema_v2_sdk import GoVarName, SDKAttribute, SDKModel
from atlas_init.cli_tf.schema_v3 import (
    TF_MODEL_NAME,
    Attribute,
    ComputedOptionalRequired,
    ListNestedAttribute,
    Resource,
    SingleNestedAttribute,
)
from atlas_init.cli_tf.schema_v3_sdk_base import (
    SDKAndSchemaAttribute,
    find_attribute,
    name_schema_struct,
    name_struct_attribute,
    schema_attributes,
)
from atlas_init.humps import pascalize

logger = logging.getLogger(__name__)


def generate_schema_to_model(resource: Resource, sdk_model: SDKModel) -> str:
    func_lines = tf_to_sdk_create_func(resource, sdk_model)
    import_urls = set()
    extend_import_urls(import_urls, func_lines)
    unformatted = "\n".join(
        [
            f"package {package_name(resource.name)}",
            "",
            *import_lines(import_urls),
            "",
            *func_lines,
        ]
    )
    return go_fmt(resource.name, unformatted)


def tf_to_sdk_create_func(resource: Resource, sdk_model: SDKModel) -> list[str]:
    lines = [
        f"func NewAtlasReq({GoVarName.CTX} context.Context, {GoVarName.INPUT} *{TF_MODEL_NAME}, {GoVarName.DIAGS} *diag.Diagnostics) *admin.{sdk_model.name} {{"
        f"  return &admin.{sdk_model.name}{{"
    ]
    nested_attributes, call_lines = convert_or_call_lines(resource, sdk_model)
    lines.extend(call_lines)
    lines.extend(
        [
            "  }"  # end struct
            "}"  # end function
        ]
    )
    lines.extend(process_nested_attributes(nested_attributes))
    return lines


_tf_to_sdk_funcs = {
    ("string", "string"): lambda tf_ref: f"{tf_ref}.ValueString()",
    ("string", "*string"): lambda tf_ref: f"{tf_ref}.ValueStringPointer()",
    ("*string", "*string"): lambda tf_ref: f"{tf_ref}.ValueStringPointer()",
    (
        "*string",
        "*time.Time",
    ): lambda tf_ref: f"conversion.StringPtrToTimePtr({tf_ref}.ValueStringPointer())",
    ("*bool", "*bool"): lambda tf_ref: f"{tf_ref}.ValueBoolPointer()",
    (
        "int64",
        "*int",
    ): lambda tf_ref: f"conversion.Int64PtrToIntPtr({tf_ref}.ValueInt64Pointer())",
    (
        "*int64",
        "*int",
    ): lambda tf_ref: f"conversion.Int64PtrToIntPtr({tf_ref}.ValueInt64Pointer())",
    ("*float64", "*float64"): lambda tf_ref: f"{tf_ref}.ValueFloat64Pointer()",
    ("*int64", "*int64"): lambda tf_ref: f"{tf_ref}.ValueInt64Pointer()",
}


def tf_to_sdk_attribute_value(
    schema_attribute: Attribute,
    sdk_attribute: SDKAttribute,
    variable_name: GoVarName = GoVarName.INPUT,
) -> str:
    key = (schema_attribute.go_type_optional, sdk_attribute.go_type)
    if key in _tf_to_sdk_funcs:
        return _tf_to_sdk_funcs[key](f"{variable_name}.{pascalize(schema_attribute.name)}")
    raise ValueError(f"Could not find conversion function for {key} for attribute: {schema_attribute.name}")


def convert_or_call_lines(
    root: Resource | Attribute,
    sdk_model: SDKModel,
    variable_name: GoVarName = GoVarName.INPUT,
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    call_lines = []
    nested_attributes: list[SDKAndSchemaAttribute] = []
    tf_attributes = schema_attributes(root)
    for sdk_attr in sorted(sdk_model.attributes.values()):
        try:
            tf_attribute = find_attribute(tf_attributes, sdk_attr.tf_name, root.name)
        except ValueError as e:
            logger.warning(e)
            continue
        if tf_attribute.computed_optional_required == ComputedOptionalRequired.computed:
            continue
        if sdk_attr.is_nested:
            call_lines.append(
                f"  {sdk_attr.struct_name}: new{sdk_attr.struct_type_name}({GoVarName.CTX}, {variable_name}.{name_struct_attribute(tf_attribute.name)}, {GoVarName.DIAGS}),"
            )
            nested_attributes.append(SDKAndSchemaAttribute(sdk_attribute=sdk_attr, schema_attribute=tf_attribute))
        elif tf_attribute.is_required:
            call_lines.append(
                f"  {sdk_attr.struct_name}: {tf_to_sdk_attribute_value(tf_attribute, sdk_attr, variable_name)},"
            )
        else:
            call_lines.append(
                f"  {sdk_attr.struct_name}: conversion.NilForUnknown({variable_name}.{tf_attribute.name_pascal}, {tf_to_sdk_attribute_value(tf_attribute, sdk_attr, variable_name)}),"
            )
    return nested_attributes, call_lines


def process_nested_attributes(
    nested_attributes: list[SDKAndSchemaAttribute],
) -> list[str]:
    lines = []
    queue = Queue()

    def add_nested_to_queue(attributes: list[SDKAndSchemaAttribute]):
        for nested in attributes:
            logger.info(f"found nested attribute: {nested.schema_attribute.name}")
            queue.put(nested)

    add_nested_to_queue(nested_attributes)
    while not queue.empty():
        sdk_attribute, schema_attribute = queue.get()
        more_nested_attributes, nested_lines = convert_nested_attribute(
            schema_attribute.nested_model, schema_attribute, sdk_attribute
        )
        lines.extend(nested_lines)
        add_nested_to_queue(more_nested_attributes)
    return lines


@singledispatch
def convert_nested_attribute(
    nested_model: object, schema_attribute: Attribute, _: SDKAttribute
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    raise NotImplementedError(f"unsupported nested attribute: {schema_attribute.name} of type {type(nested_model)}")


@convert_nested_attribute.register
def _convert_single_nested_attribute(
    _: SingleNestedAttribute,
    schema_attribute: Attribute,
    sdk_attribute: SDKAttribute,
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    sdk_model = sdk_attribute.as_sdk_model()
    lines: list[str] = [
        f"func new{sdk_model.name}(ctx context.Context, {GoVarName.INPUT} types.Object, diags *diag.Diagnostics) *admin.{sdk_model.name} {{",
        f"  var resp *admin.{sdk_model.name}",
        f"  if {GoVarName.INPUT}.IsUnknown() || {GoVarName.INPUT}.IsNull() {{",
        "    return resp",
        "  }",
        f"  {GoVarName.ITEM} := &{name_schema_struct(schema_attribute.name)}{{}}",
        f"  if localDiags := {GoVarName.INPUT}.As({GoVarName.CTX}, {GoVarName.ITEM}, basetypes.ObjectAsOptions{{}}); len(localDiags) > 0 {{",
        f"    {GoVarName.DIAGS}.Append(localDiags...)",
        "    return resp",
        "  }",
        f"  return &admin.{sdk_model.name}{{",
    ]
    nested_attributes, call_lines = convert_or_call_lines(schema_attribute, sdk_model, GoVarName.ITEM)
    lines.extend([*call_lines, "  }", "}"])  # end struct  # end function
    return nested_attributes, lines


@convert_nested_attribute.register
def _convert_list_nested_attriute(
    _: ListNestedAttribute,
    schema_attribute: Attribute,
    sdk_attribute: SDKAttribute,
) -> tuple[list[SDKAndSchemaAttribute], list[str]]:
    sdk_model = sdk_attribute.as_sdk_model()
    lines: list[str] = [
        f"func new{sdk_model.name}(ctx context.Context, {GoVarName.INPUT} types.List, diags *diag.Diagnostics) *[]admin.{sdk_model.name} {{",
        f"  if {GoVarName.INPUT}.IsUnknown() || {GoVarName.INPUT}.IsNull() {{",
        "    return nil",
        "  }",
        f"  {GoVarName.ELEMENTS} := make([]{name_schema_struct(schema_attribute.name)}, len({GoVarName.INPUT}.Elements()))",
        f"  if localDiags := {GoVarName.INPUT}.ElementsAs({GoVarName.CTX}, &{GoVarName.ELEMENTS}, false); len(localDiags) > 0 {{",
        f"    {GoVarName.DIAGS}.Append(localDiags...)",
        "    return nil",
        "  }",
        f"  {GoVarName.RESP} := make([]admin.{sdk_model.name}, len({GoVarName.INPUT}.Elements()))",
        f"  for i := range {GoVarName.ELEMENTS} {{",
        f"    {GoVarName.ITEM} := &{GoVarName.ELEMENTS}[i]",
        f"    resp[i] = admin.{sdk_model.name}{{",
    ]
    nested_attributes, call_lines = convert_or_call_lines(schema_attribute, sdk_model, GoVarName.ITEM)
    lines.extend(
        [
            *[f"    {line}" for line in call_lines],
            "    }",  # end struct
            "  }",  # end loop
            "  return &resp",
            "}",  # end function
        ]
    )
    return nested_attributes, lines



import logging
from pathlib import Path

import typer

logger = logging.getLogger(__name__)
SPLIT_STR = "mongodbatlas: "


def remove_prefix(line: str) -> str:
    """
    >>> remove_prefix(
    ...     "2025-02-14T15:47:14.157Z [DEBUG] provider.terraform-provider-mongodbatlas: {"
    ... )
    '{'
    >>> remove_prefix(
    ...     '2025-02-14T15:47:14.158Z [DEBUG] provider.terraform-provider-mongodbatlas:  "biConnector": {'
    ... )
    ' "biConnector": {'
    """
    return line if SPLIT_STR not in line else line.split(SPLIT_STR, 1)[1]


def log_clean(log_path: str = typer.Argument(..., help="Path to the log file")):
    log_path_parsed = Path(log_path)
    assert log_path_parsed.exists(), f"file not found: {log_path}"
    new_lines = [remove_prefix(line) for line in log_path_parsed.read_text().splitlines()]
    log_path_parsed.write_text("\n".join(new_lines))
    logger.info(f"cleaned log file: {log_path}")



import logging

from rich import prompt

logger = logging.getLogger(__name__)


def confirm(prompt_text: str, *, is_interactive: bool, default: bool) -> bool:
    if not is_interactive:
        logger.warning(f"non-interactive prompt: {prompt_text}, using default={default}")
        return default
    return prompt.Confirm(prompt_text)()



import logging
import sys
from pathlib import Path

import typer
from zero_3rdparty.file_utils import clean_dir

from atlas_init.cli_args import option_sdk_repo_path
from atlas_init.cli_helper.run import (
    run_binary_command_is_ok,
    run_command_exit_on_failure,
)
from atlas_init.cli_tf.ci_tests import ci_tests
from atlas_init.cli_tf.changelog import convert_to_changelog
from atlas_init.cli_tf.example_update import update_example_cmd
from atlas_init.cli_tf.log_clean import log_clean
from atlas_init.cli_tf.mock_tf_log import mock_tf_log_cmd
from atlas_init.cli_tf.schema import (
    dump_generator_config,
    parse_py_terraform_schema,
    update_provider_code_spec,
)
from atlas_init.cli_tf.schema_inspection import log_optional_only
from atlas_init.cli_tf.schema_v2 import (
    generate_resource_go_resource_schema,
    parse_schema,
)
from atlas_init.cli_tf.openapi import add_api_spec_info
from atlas_init.cli_tf.schema_v2_sdk import generate_model_go, parse_sdk_model
from atlas_init.repos.go_sdk import download_admin_api
from atlas_init.repos.path import Repo, current_repo_path
from atlas_init.settings.env_vars import init_settings
from atlas_init.settings.interactive import confirm

app = typer.Typer(no_args_is_help=True)
app.command(name="mock-tf-log")(mock_tf_log_cmd)
app.command(name="example-update")(update_example_cmd)
app.command(name="log-clean")(log_clean)
app.command(name="ci-tests")(ci_tests)
logger = logging.getLogger(__name__)


@app.command()
def schema(
    branch: str = typer.Option("main", "-b", "--branch"),
):
    settings = init_settings()
    schema_out_path = settings.schema_out_path_computed
    schema_out_path.mkdir(exist_ok=True)

    schema_parsed = parse_py_terraform_schema(settings.atlas_init_tf_schema_config_path)
    generator_config = dump_generator_config(schema_parsed)
    generator_config_path = schema_out_path / "generator_config.yaml"
    generator_config_path.write_text(generator_config)
    provider_code_spec_path = schema_out_path / "provider-code-spec.json"
    admin_api_path = schema_out_path / "admin_api.yaml"

    if admin_api_path.exists():
        logger.warning(f"using existing admin api @ {admin_api_path}")
    else:
        download_admin_api(admin_api_path, branch=branch)

    if not run_binary_command_is_ok(
        cwd=schema_out_path,
        binary_name="tfplugingen-openapi",
        command=f"generate --config {generator_config_path.name} --output {provider_code_spec_path.name} {admin_api_path.name}",
        logger=logger,
    ):
        logger.critical("failed to generate spec")
        sys.exit(1)
    new_provider_spec = update_provider_code_spec(schema_parsed, provider_code_spec_path)
    provider_code_spec_path.write_text(new_provider_spec)
    logger.info(f"updated {provider_code_spec_path.name} ✅ ")

    go_code_output = schema_out_path / "internal"
    if go_code_output.exists():
        logger.warning(f"cleaning go code dir: {go_code_output}")
        clean_dir(go_code_output, recreate=True)

    if not run_binary_command_is_ok(
        cwd=schema_out_path,
        binary_name="tfplugingen-framework",
        command=f"generate all --input ./{provider_code_spec_path.name} --output {go_code_output.name}",
        logger=logger,
    ):
        logger.critical("failed to generate plugin schema")
        sys.exit(1)

    logger.info(f"new files generated to {go_code_output} ✅")
    for go_file in sorted(go_code_output.rglob("*.go")):
        logger.info(f"new file @ '{go_file}'")


@app.command()
def schema_optional_only():
    repo_path = current_repo_path(Repo.TF)
    log_optional_only(repo_path)


@app.command()
def changelog(
    pr: str = typer.Argument("", help="the PR number, will read the file in .changelog/$pr_input.txt"),
    delete_input: bool = typer.Option(False, "-d", "--delete-input"),
):
    repo_path = current_repo_path(Repo.TF)
    changelog_input_path = repo_path / f".changelog/{pr}_input.txt"
    if not changelog_input_path.exists():
        logger.critical(f"no file @ {changelog_input_path}")
        raise typer.Abort
    changes_in = changelog_input_path.read_text()
    logger.info(f"will generate changelog to {changelog_input_path} based on changes:\n{changes_in}")
    changes_out = convert_to_changelog(changes_in)
    changelog_path = repo_path / f".changelog/{pr}.txt"
    changelog_path.write_text(changes_out)
    logger.info(f"updated file ✅ \n{changes_in}\n--> TO:\n{changes_out} ")
    if delete_input:
        logger.warning(f"deleting input file {changelog_input_path}")
        changelog_input_path.unlink()


@app.command()
def example_gen(
    in_path: Path = typer.Argument(..., help="Path to the latest code"),
    out_path: Path = typer.Argument("", help="Output path (empty will use input path)"),
):
    out_path = out_path or in_path  # type: ignore
    assert in_path.is_dir(), f"path not found: {in_path}"
    assert out_path.is_dir(), f"path not found: {out_path}"
    run_command_exit_on_failure("terraform fmt", cwd=in_path, logger=logger)
    if in_path == out_path:
        logger.warning(f"will overwrite/change files in {out_path}")
    else:
        logger.info(f"will use from {in_path} -> {out_path}")
    from zero_3rdparty import file_utils

    for path, rel_path in file_utils.iter_paths_and_relative(in_path, "*.tf", "*.sh", "*.py", "*.md", rglob=False):
        dest_path = out_path / rel_path
        file_utils.copy(path, dest_path, clean_dest=False)


@app.command()
def schema2(
    resource: str = typer.Argument(
        "",
        help="the resource name to generate the schema for. Must exist in the schema. E.g., 'stream_processor'",
    ),
    branch: str = typer.Option("main", "-b", "--branch", help="the branch for downloading openapi spec"),
    admin_api_path: Path = typer.Option(
        "", "-a", "--admin-api-path", help="the path to store/download the openapi spec"
    ),
    config_path: Path = typer.Option("", "-c", "--config", help="the path to the SchemaV2 config"),
    replace: bool = typer.Option(False, "-r", "--replace", help="replace the existing schema file"),
    sdk_repo_path_str: str = option_sdk_repo_path,
):
    repo_path = current_repo_path(Repo.TF)
    config_path = config_path or repo_path / "schema_v2.yaml"
    admin_api_path = admin_api_path or repo_path / "admin_api.yaml"
    if admin_api_path.exists():
        logger.info(f"using existing admin api @ {admin_api_path}")
    else:
        download_admin_api(admin_api_path, branch=branch)
    schema = parse_schema(config_path)
    logger.info("adding api spec info to schema")
    add_api_spec_info(schema, admin_api_path, minimal_refs=True)
    go_old = repo_path / f"internal/service/{resource.replace('_', '')}/resource_schema.go"
    if not go_old.exists():
        if confirm(
            f"no file found @ {go_old}, ok to create it?",
            is_interactive=True,
            default=True,
        ):
            go_old.parent.mkdir(exist_ok=True, parents=True)
        else:
            logger.critical(f"no file found @ {go_old}")
            raise typer.Abort
    if replace:
        logger.warning(f"replacing existing schema @ {go_old}")
        go_new = go_old
    else:
        go_new = go_old.with_name("resource_schema_gen.go")
    gen_src = generate_resource_go_resource_schema(schema, resource)
    go_new.write_text(gen_src)
    logger.info(f"generated new schema @ {go_new} ✅")

    resource_schema = schema.resources[resource]
    if conversion_config := resource_schema.conversion:
        if not confirm(
            f"resource {resource} has conversion, ok to generate conversion functions?",
            is_interactive=True,
            default=True,
        ):
            logger.info("skipping conversion functions")
            return
        logger.info("generating conversion functions")
        if not sdk_repo_path_str:
            logger.critical("must provide sdk repo path when generating conversion functions")
            raise typer.Abort
        sdk_repo_path = Path(sdk_repo_path_str)
        if not sdk_repo_path.exists():
            logger.critical(f"no sdk repo found @ {sdk_repo_path}")
            raise typer.Abort
        for sdk_start_ref in conversion_config.sdk_start_refs:
            sdk_model = parse_sdk_model(sdk_repo_path, sdk_start_ref.name)
            go_conversion_src = generate_model_go(schema, resource_schema, sdk_model)
            go_conversion_path = go_old.with_name("model.go")
            go_conversion_path.write_text(go_conversion_src)



from __future__ import annotations
import logging
from concurrent.futures import Future
from contextlib import suppress
from datetime import datetime
from typing import ClassVar
from ask_shell import ShellRun, confirm, kill, run, run_and_wait
from ask_shell.models import ShellRunEventT, ShellRunStdOutput
from zero_3rdparty import str_utils
from zero_3rdparty.file_utils import copy, ensure_parents_write_text
from zero_3rdparty.future import chain_future
from atlas_init.settings.env_vars import AtlasInitSettings
from pathlib import Path
from model_lib import Event

logger = logging.getLogger(__name__)


class MonthlyReportPaths(Event):
    summary_path: Path
    error_only_path: Path
    details_dir: Path
    summary_name: str
    daily_path: Path

    ERROR_ONLY_SUFFIX: ClassVar[str] = "_error-only.md"
    DAILY_SUFFIX: ClassVar[str] = "_daily.md"

    def export_to_dir(self, out_dir: Path) -> None:
        for path in [self.summary_path, self.error_only_path, self.daily_path]:
            if path.exists():
                ensure_parents_write_text(out_dir / path.name, path.read_text())
        if self.details_dir.exists():
            copy(self.details_dir, out_dir / self.details_dir.name, clean_dest=True)

    @classmethod
    def from_settings(cls, settings: AtlasInitSettings, summary_name: str) -> MonthlyReportPaths:
        summary_path = settings.github_ci_summary_dir / str_utils.ensure_suffix(summary_name, ".md")
        return cls(
            summary_path=summary_path,
            error_only_path=settings.github_ci_summary_dir
            / str_utils.ensure_suffix(summary_name, MonthlyReportPaths.ERROR_ONLY_SUFFIX),
            details_dir=settings.github_ci_summary_details_path(summary_name, "dummy").parent,
            summary_name=summary_name,
            daily_path=settings.github_ci_summary_dir / f"{summary_path.stem}{MonthlyReportPaths.DAILY_SUFFIX}",
        )


CI_TESTS_DIR_NAME = "ci-tests"
MKDOCS_SERVE_TIMEOUT = 120
MKDOCS_SERVE_URL = "http://127.0.0.1:8000"


def export_ci_tests_markdown_to_html(settings: AtlasInitSettings, report_paths: MonthlyReportPaths) -> None:
    html_out = settings.atlas_init_static_html_path
    if not html_out or not html_out.exists():
        return
    ci_tests_dir = html_out / CI_TESTS_DIR_NAME
    docs_out_dir = ci_tests_dir / "docs"
    report_paths.export_to_dir(docs_out_dir)
    index_md_content = create_index_md(docs_out_dir)
    ensure_parents_write_text(docs_out_dir / "index.md", index_md_content)
    server_url, run_event = start_mkdocs_serve(ci_tests_dir)
    try:
        if confirm(f"do you want to open the html docs? {server_url}", default=False):
            run_and_wait(f'open -a "Google Chrome" {server_url}')
        if confirm("Finished testing html docs?", default=False):
            pass
    except BaseException as e:
        raise e
    finally:
        kill(run_event, reason="Done with html docs check")
    if confirm("Are docs ok to build and push?", default=False):
        build_and_push(ci_tests_dir, report_paths.summary_name)


def create_index_md(docs_out_dir: Path) -> str:
    """
    tree -L 1 docs
    docs
    ├── 2025-06-26_details
    ├── 2025-06-26_.md
    ├── 2025-06-26.md
    ├── 2025-06-26_error-only.md
    ├── index.md
    ├── javascript
    └── stylesheets
    """
    md_files = {f.name: f for f in docs_out_dir.glob("*.md") if f.name != "index.md"}
    parsed_dates = []
    for md_file in md_files.values():
        with suppress(ValueError):
            parsed_dates.append(datetime.strptime(md_file.stem, "%Y-%m-%d"))
    parsed_dates.sort(reverse=True)

    def date_row(date: datetime) -> str:
        summary_filename = f"{date.strftime('%Y-%m-%d')}.md"
        error_only_filename = f"{date.strftime('%Y-%m-%d')}{MonthlyReportPaths.ERROR_ONLY_SUFFIX}"
        daily_filename = f"{date.strftime('%Y-%m-%d')}{MonthlyReportPaths.DAILY_SUFFIX}"
        line_links = [f"[{date.strftime('%Y-%m-%d')}](./{summary_filename})"]
        if error_only_filename in md_files:
            line_links.append(f"[{date.strftime('%Y-%m-%d')} Error Only](./{error_only_filename})")
        if daily_filename in md_files:
            line_links.append(f"[{date.strftime('%Y-%m-%d')} Daily Errors](./{daily_filename})")
        return f"- {', '.join(line_links)}"

    md_content = [
        "# Welcome to CI Tests",
        "",
        *[date_row(dt) for dt in parsed_dates],
        "",
    ]
    return "\n".join(md_content)


def start_mkdocs_serve(ci_tests_dir: Path) -> tuple[str, ShellRun]:
    future = Future()

    def on_message(event: ShellRunEventT) -> bool:
        match event:
            case ShellRunStdOutput(_, content) if f"Serving on {MKDOCS_SERVE_URL}" in content:
                logger.info(f"Docs server ready @ {MKDOCS_SERVE_URL}")
                future.set_result(None)
                return True
        return False

    run_event = run(
        "uv run mkdocs serve", cwd=ci_tests_dir, message_callbacks=[on_message], print_prefix="mkdocs serve"
    )
    chain_future(run_event._complete_flag, future)
    try:
        future.result(timeout=MKDOCS_SERVE_TIMEOUT)
    except BaseException as e:
        kill(run_event, reason=f"Failed to start mkdocs serve, timeout after {MKDOCS_SERVE_TIMEOUT} seconds")
        raise e
    return MKDOCS_SERVE_URL, run_event


def build_and_push(ci_tests_dir: Path, summary_name: str) -> None:
    run_and_wait("uv run mkdocs build", cwd=ci_tests_dir, print_prefix="build")
    run_and_wait("git add .", cwd=ci_tests_dir, print_prefix="add")
    run_and_wait(f"git commit -m 'update ci tests {summary_name}'", cwd=ci_tests_dir, print_prefix="commit")
    run_and_wait("git push", cwd=ci_tests_dir, print_prefix="push")



import contextlib
import logging
import re
from collections.abc import Iterable
from pathlib import Path

import stringcase
from model_lib import Entity, dump, parse_model, parse_payload
from pydantic import ConfigDict, Field, ValidationError
from zero_3rdparty import file_utils

logger = logging.getLogger(__name__)


def create_sample_file_from_input(samples_dir: Path, log_group_name: str, inputs_file: Path) -> Path:
    resource_state = parse_payload(inputs_file)
    assert isinstance(resource_state, dict), f"input file with not a dict {resource_state}"
    samples_file = samples_dir / inputs_file.name
    if inputs_file.name.endswith("_create.json"):
        return create_sample_file(samples_file, log_group_name, resource_state)
    if inputs_file.name.endswith("_update.json"):
        prev_state_path = inputs_file.parent / inputs_file.name.replace("_update.json", "_create.json")
        prev_state: dict = parse_payload(prev_state_path)  # type: ignore
        return create_sample_file(
            samples_file,
            log_group_name,
            resource_state,
            prev_resource_state=prev_state,
        )
    raise ValueError(f"unexpected input file {inputs_file}")


def create_sample_file(
    samples_file: Path,
    log_group_name: str,
    resource_state: dict,
    prev_resource_state: dict | None = None,
) -> Path:
    logger.info(f"adding sample @ {samples_file}")
    assert isinstance(resource_state, dict)
    new_json = dump(
        {
            "providerLogGroupName": log_group_name,
            "previousResourceState": prev_resource_state or {},
            "desiredResourceState": resource_state,
        },
        "pretty_json",
    )
    file_utils.ensure_parents_write_text(samples_file, new_json)
    return samples_file


CamelAlias = ConfigDict(alias_generator=stringcase.camelcase, populate_by_name=True)


class CfnSchema(Entity):
    model_config = CamelAlias

    description: str
    type_name: str = Field(pattern=r"^[a-zA-Z0-9]{2,64}::[a-zA-Z0-9]{2,64}::[a-zA-Z0-9]{2,64}$")


def iterate_schemas(resource_root: Path) -> Iterable[tuple[Path, CfnSchema]]:
    for path in resource_root.rglob("*.json"):
        if path.parent.parent != resource_root:
            continue
        with contextlib.suppress(ValidationError):
            yield path, parse_model(path, t=CfnSchema)


_md_patterns = [
    re.compile(r"\[[^\]]+\]\([^\)]+\)"),
    re.compile(r"`\S+`"),
]


def has_md_link(text: str) -> bool:
    return any(bool(pattern.findall(text)) for pattern in _md_patterns)


def default_log_group_name(resource_name: str) -> str:
    return f"mongodb-atlas-{resource_name}-logs"



import logging

import typer

from atlas_init.cli_helper.go import GoEnvVars, GoTestCaptureMode, GoTestMode, GoTestResult, run_go_tests
from atlas_init.cli_tf.mock_tf_log import MockTFLog, mock_tf_log, resolve_admin_api_path
from atlas_init.repos.path import Repo, current_repo, current_repo_path
from atlas_init.settings.env_vars import active_suites, init_settings
from atlas_init.typer_app import app_command

logger = logging.getLogger(__name__)


@app_command()
def go_test(
    mode: str = typer.Option("package", "-m", "--mode", help="package|individual or a prefix"),
    dry_run: bool = typer.Option(False, help="only log out the commands to be run"),
    timeout_minutes: int = typer.Option(300, "-t", "--timeout", help="timeout in minutes"),
    concurrent_runs: int = typer.Option(20, "-c", "--concurrent", help="number of concurrent runs"),
    re_run: bool = typer.Option(False, "-r", "--re-run", help="re-run the tests if the log already exist"),
    export_mock_tf_log: bool = typer.Option(False, "-e", "--export", help="export the mock-tf-log"),
    export_mock_tf_log_verbose: bool = typer.Option(
        False, "--export-verbose", help="log roundtrips when exporting the mock-tf-log"
    ),
    env_method: GoEnvVars = typer.Option(GoEnvVars.manual, "--env"),
    names: list[str] = typer.Option(
        ...,
        "-n",
        "--names",
        default_factory=list,
        help="run only the tests with these names",
    ),
    capture_mode: GoTestCaptureMode = typer.Option(GoTestCaptureMode.capture, "--capture"),
    use_old_schema: bool = typer.Option(False, "--old-schema", help="use the old schema for the tests"),
):
    if export_mock_tf_log and mode != GoTestMode.individual:
        err_msg = "exporting mock-tf-log is only supported for individual tests"
        raise ValueError(err_msg)
    settings = init_settings()
    suites = active_suites(settings)
    sorted_suites = sorted(suite.name for suite in suites)
    logger.info(f"running go tests for {len(suites)} test-suites: {sorted_suites}")
    results: GoTestResult | None = None
    match current_repo():
        case Repo.CFN:
            raise NotImplementedError
        case Repo.TF:
            repo_path = current_repo_path()
            results = run_go_tests(
                repo_path,
                settings,
                suites,
                mode,
                dry_run=dry_run,
                timeout_minutes=timeout_minutes,
                concurrent_runs=concurrent_runs,
                re_run=re_run,
                env_vars=env_method,
                names=set(names),
                capture_mode=capture_mode,
                use_old_schema=use_old_schema,
            )
        case _:
            raise NotImplementedError
    if results is None:
        error_msg = "no results found"
        raise ValueError(error_msg)
    if export_mock_tf_log:
        _export_mock_tf_logs(results, export_mock_tf_log_verbose)
    # use the test_results: dict[str, list[GoTestRun]]
    # TODO: create_detailed_summary()


def _export_mock_tf_logs(results: GoTestResult, verbose: bool):
    package_paths = results.test_name_package_path
    admin_api_path = resolve_admin_api_path("", sdk_branch="main", admin_api_path="")
    for test_name, runs in results.runs.items():
        package_path = package_paths.get(test_name)
        if package_path is None:
            logger.warning(f"no package path found for test_name={test_name}")
            continue
        assert len(runs) == 1, f"expected only 1 run for test_name={test_name}, got {len(runs)}"
        run = runs[0]
        tpf_package_path = package_path.with_name(f"{package_path.name}tpf") / "testdata"
        default_package_path = package_path / "testdata"
        if not tpf_package_path.exists():
            logger.warning(
                f"tpf_package_path={tpf_package_path} doesn't exist, adding mocked data to {default_package_path}"
            )
            tpf_package_path = default_package_path
        tf_log_path = run.log_path
        assert tf_log_path, f"test didn't set tf_log_path: {test_name}"
        if test_name in results.failure_names:
            logger.warning(f"test_name={test_name} failed, not exporting mock-tf-log")
            continue
        req = MockTFLog(
            log_path=tf_log_path,
            output_dir=tpf_package_path,
            admin_api_path=admin_api_path,
            package_name=package_path.name,
            log_diff_roundtrips=verbose,
        )
        mocked_yaml = mock_tf_log(req)
        logger.info(f"mocked TestConfig saved to {mocked_yaml}")



from __future__ import annotations

from enum import Enum, StrEnum
from typing import TypeAlias

from pydantic import BaseModel, ConfigDict, Field

from atlas_init.humps import pascalize


def lowercase_no_snake(name: str) -> str:
    return name.lower().replace("_", "")


def default_config_dict() -> ConfigDict:
    return ConfigDict(populate_by_name=True, alias_generator=lowercase_no_snake)


class BaseModelLocal(BaseModel):
    model_config = default_config_dict()


class ElemType(int, Enum):
    BOOL = 0
    FLOAT64 = 1
    INT64 = 2
    NUMBER = 3
    STRING = 4
    UNKNOWN = 5


class CustomDefault(BaseModelLocal):
    definition: str
    imports: list[str]


SnakeCaseString = str


class ComputedOptionalRequired(StrEnum):
    computed = "computed"
    computed_optional = "computed_optional"
    optional = "optional"
    required = "required"
    unset = ""


class BoolAttribute(BaseModelLocal):
    default: bool | None = None


class Float64Attribute(BaseModelLocal):
    default: float | None = None


class Int64Attribute(BaseModelLocal):
    default: int | None = None


class MapAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    element_type: ElemType


class NestedAttributeObject(BaseModelLocal):
    attributes: list[Attribute]


class MapNestedAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    nested_object: NestedAttributeObject


class NumberAttribute(BaseModelLocal):
    default: CustomDefault | None = None


class SetAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    element_type: ElemType


class SetNestedAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    nested_object: NestedAttributeObject


class SingleNestedAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    nested_object: NestedAttributeObject


class StringAttribute(BaseModelLocal):
    default: str | None = None


class ListAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    element_type: ElemType


class ListNestedAttribute(BaseModelLocal):
    default: CustomDefault | None = None
    nested_object: NestedAttributeObject


class Operation(int, Enum):
    CREATE = 0
    UPDATE = 1
    READ = 2
    DELETE = 3


class TimeoutsAttribute(BaseModelLocal):
    configurable_timeouts: list[Operation] = Field(default_factory=list, alias="configurabletimeouts")


NestedModelT: TypeAlias = SetNestedAttribute | ListNestedAttribute | MapNestedAttribute | SingleNestedAttribute


class Attribute(BaseModelLocal):
    list: ListAttribute | None = None
    float64: Float64Attribute | None = None
    string: StringAttribute | None = None
    bool: BoolAttribute | None = None
    map: MapAttribute | None = None
    number: NumberAttribute | None = None
    set: SetAttribute | None = None
    int64: Int64Attribute | None = None

    set_nested: SetNestedAttribute | None = None
    list_nested: ListNestedAttribute | None = None
    map_nested: MapNestedAttribute | None = None
    single_nested: SingleNestedAttribute | None = None

    timeouts: TimeoutsAttribute | None = None
    description: str | None = None
    name: SnakeCaseString
    deprecation_message: str | None = None
    sensitive: bool = False

    computed_optional_required: ComputedOptionalRequired | None = Field(default=None, alias="computedoptionalrequired")

    @property
    def name_pascal(self):
        return pascalize(self.name)

    @property
    def is_nested(self) -> bool:
        return any([self.single_nested, self.list_nested, self.map_nested, self.set_nested])

    @property
    def nested_model(self) -> NestedModelT:
        assert self.is_nested
        return self.single_nested or self.list_nested or self.map_nested or self.set_nested  # type: ignore

    @property
    def is_attribute(self) -> bool:
        return self.computed_optional_required != ComputedOptionalRequired.unset

    @property
    def nested_attributes(self) -> list[Attribute]:
        return self.nested_model.nested_object.attributes

    @property
    def is_optional(self) -> bool:
        return self.computed_optional_required in {
            ComputedOptionalRequired.optional,
            ComputedOptionalRequired.computed_optional,
            ComputedOptionalRequired.computed,
        }

    @property
    def is_required(self) -> bool:
        return self.computed_optional_required == ComputedOptionalRequired.required

    @property
    def go_type(self) -> str:
        assert not self.is_nested
        if self.bool:
            return "bool"
        if self.float64:
            return "float64"
        if self.int64:
            return "int64"
        if self.number:
            return "number"
        if self.string:
            return "string"
        if self.list:
            raise NotImplementedError(f"list {self.name}")
        if self.map:
            if self.map.element_type == ElemType.STRING:
                return "map[string]string"
            raise NotImplementedError(f"map {self.name}")
        if self.set:
            raise NotImplementedError(f"set {self.name}")
        raise NotImplementedError(f"unknown type: {self.name}")

    @property
    def go_type_optional(self) -> str:
        assert not self.is_nested, "should find go_type_optional for a nested attribute"
        return f"*{self.go_type}" if self.is_optional else self.go_type


class Schema(BaseModelLocal):
    description: str | None = None
    deprecation_message: str | None = None
    attributes: list[Attribute]


class Resource(BaseModelLocal):
    local_schema: Schema = Field(alias="schema")
    name: SnakeCaseString

    @property
    def use_timeout(self) -> bool:
        return any(a.timeouts for a in self.local_schema.attributes)


ResourceSchemaV3 = Resource
TF_MODEL_NAME = "TFModel"



from ask_shell import configure_logging
from typer import Typer


def typer_main():
    from atlas_init.sdk_ext import go

    app = Typer(
        name="sdk-ext",
        help="SDK extension commands for Atlas Init",
    )
    app.command(name="go")(go.go)
    configure_logging(app)
    app()


if __name__ == "__main__":
    typer_main()



# pyright: reportIncompatibleMethodOverride=none
from collections import defaultdict
from collections.abc import Iterable
from functools import total_ordering
from pathlib import Path
from typing import Literal, TypeAlias

from model_lib import Entity, Event
from pydantic import Field, model_validator
from zero_3rdparty import iter_utils

from atlas_init.cli_tf.schema_go_parser import parse_schema_functions
from atlas_init.cli_tf.schema_table_models import TFSchemaAttribute, TFSchemaTableColumn
from atlas_init.settings.path import default_factory_cwd


def default_table_columns() -> list[TFSchemaTableColumn]:
    return [TFSchemaTableColumn.Computability]


def file_name_path(file: str) -> tuple[str, Path]:
    if ":" in file:
        file, path = file.split(":", 1)
        return file, Path(path)
    path = Path(file)
    return f"{path.parent.name}/{path.stem}"[:20], path


@total_ordering
class TFSchemaSrc(Event):
    name: str
    file_path: Path | None = None
    url: str = ""

    @model_validator(mode="after")
    def validate(self):
        assert self.file_path or self.url, "must provide file path or url"
        if self.file_path:
            assert self.file_path.exists(), f"file does not exist for {self.name}: {self.file_path}"
        return self

    def __lt__(self, other) -> bool:
        if not isinstance(other, TFSchemaSrc):
            raise TypeError
        return self.name < other.name

    def go_code(self) -> str:
        if path := self.file_path:
            return path.read_text()
        raise NotImplementedError


TableOutputFormats: TypeAlias = Literal["md"]


class TFSchemaTableInput(Entity):
    sources: list[TFSchemaSrc] = Field(default_factory=list)
    output_format: TableOutputFormats = "md"
    output_path: Path = Field(default_factory=default_factory_cwd("schema_table.md"))
    columns: list[TFSchemaTableColumn] = Field(default_factory=default_table_columns)
    explode_rows: bool = False

    @model_validator(mode="after")
    def validate(self):
        assert self.columns, "must provide at least 1 column"
        self.columns = sorted(self.columns)
        assert self.sources, "must provide at least 1 source"
        self.sources = sorted(self.sources)
        assert len(self.sources) == len(set(self.sources)), f"duplicate source names: {self.source_names}"
        return self

    @property
    def source_names(self) -> list[str]:
        return [s.name for s in self.sources]

    def headers(self) -> list[str]:
        return ["Attribute Name"] + [f"{name}-{col}" for name in self.source_names for col in self.columns]


@total_ordering
class TFSchemaTableData(Event):
    source: TFSchemaSrc
    schema_path: str = ""  # e.g., "" is root, "replication_specs.region_config"
    attributes: list[TFSchemaAttribute] = Field(default_factory=list)

    @property
    def id(self) -> str:
        return f"{self.schema_path}:{self.source.name}"

    def __lt__(self, other) -> bool:
        if not isinstance(other, TFSchemaTableData):
            raise TypeError
        return self.id < other.id


def sorted_schema_paths(schema_paths: Iterable[str]) -> list[str]:
    return sorted(schema_paths, key=lambda x: (x.count("."), x.split(".")[-1]))


class RawTable(Event):
    columns: list[str]
    rows: list[list[str]]


def merge_tables(config: TFSchemaTableInput, schema_path: str, tables: list[TFSchemaTableData]) -> RawTable:
    if schema_path != "":
        raise NotImplementedError
    columns = config.headers()
    if len(tables) > 1:
        err_msg = "only 1 table per schema path supported"
        raise NotImplementedError(err_msg)
    table = tables[0]
    rows = [[attr.absolute_attribute_path, *attr.row(config.columns)] for attr in table.attributes]
    return RawTable(columns=columns, rows=rows)


def format_table(table: RawTable, table_format: TableOutputFormats) -> list[str]:
    # sourcery skip: merge-list-append
    assert table_format == "md", "only markdown format supported"
    lines = []
    lines.append("|".join(table.columns))
    lines.append("|".join(["---"] * len(table.columns)))
    lines.extend("|".join(row) for row in table.rows)
    return lines


def explode_attributes(attributes: list[TFSchemaAttribute]) -> list[TFSchemaAttribute]:
    return sorted(iter_utils.flat_map(attr.explode() for attr in attributes))


def schema_table(config: TFSchemaTableInput) -> str:
    path_tables: dict[str, list[TFSchemaTableData]] = defaultdict(list)
    for source in config.sources:
        go_code = source.go_code()
        attributes, functions = parse_schema_functions(go_code)
        if config.explode_rows:
            attributes = explode_attributes(attributes)
        schema_path = ""  # using only root for now
        path_tables[schema_path].append(
            TFSchemaTableData(source=source, attributes=attributes, schema_path=schema_path)
        )
    output_lines = []
    for schema_path in sorted_schema_paths(path_tables.keys()):
        tables = path_tables[schema_path]
        table = merge_tables(config, schema_path, tables)
        output_lines.extend(["", f"## {schema_path or 'Root'}", ""])
        output_lines.extend(format_table(table, table_format=config.output_format))
    return "\n".join(output_lines)



from __future__ import annotations

from dataclasses import dataclass, field
import logging
from typing import TypeAlias

from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase, AsyncIOMotorCollection
from pymongo import IndexModel
from pymongo.errors import DuplicateKeyError
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

from atlas_init.cli_tf.go_test_run import GoTestRun
from atlas_init.cli_tf.go_test_tf_error import GoTestErrorClassification
from atlas_init.crud.mongo_utils import index_dec


logger = logging.getLogger(__name__)


@dataclass
class CollectionConfig:
    name: str = ""  # uses the class name by default
    indexes: list[IndexModel] = field(default_factory=list)


CollectionConfigsT: TypeAlias = dict[type, CollectionConfig]


def default_document_models() -> CollectionConfigsT:
    return {
        GoTestErrorClassification: CollectionConfig(
            indexes=[index_dec("ts"), IndexModel(["error_class"]), IndexModel(["test_name"])]
        ),
        GoTestRun: CollectionConfig(indexes=[index_dec("ts"), IndexModel(["branch"]), IndexModel(["status"])]),
    }


_collections = {}


def get_collection(model: type) -> AsyncIOMotorCollection:
    col = _collections.get(model)
    if col is not None:
        return col
    raise ValueError(f"Collection for model {model.__name__} is not initialized. Call init_mongo first.")


def get_db(mongo_url: str, db_name: str) -> AsyncIOMotorDatabase:
    client = AsyncIOMotorClient(mongo_url)
    return client.get_database(db_name)


async def init_mongo(
    mongo_url: str, db_name: str, clean_collections: bool = False, document_models: CollectionConfigsT | None = None
) -> None:
    db = get_db(mongo_url, db_name)
    document_models = document_models or default_document_models()
    for model, cfg in document_models.items():
        name = cfg.name or model.__name__
        col = await ensure_collection_exist(db, name, cfg.indexes, clean_collections)
        _collections[model] = col

    if clean_collections:
        logger.info(f"MongoDB collections in '{db_name}' have been cleaned.")


async def ensure_collection_exist(
    db: AsyncIOMotorDatabase,
    name: str,
    indexes: list[IndexModel] | None = None,
    clean_collection: bool = False,
) -> AsyncIOMotorCollection:
    existing = await db.list_collection_names()
    if clean_collection and name in existing:
        await db.drop_collection(name)
        existing.remove(name)

    if name not in existing:
        await db.create_collection(name)

    if indexes:
        # always (re-)create indexes after new creation or drop
        await db[name].create_indexes(indexes)

    logger.debug(f"mongo collection {name!r} is ready")
    return db[name]


def duplicate_key_pattern(error: DuplicateKeyError) -> str | None:
    details: dict = error.details  # type: ignore
    name_violator = details.get("keyPattern", {})
    if not name_violator:
        return None
    name, _ = name_violator.popitem()
    return name


class CollectionNotEmptyError(Exception):
    def __init__(self, collection_name: str):
        super().__init__(f"Collection '{collection_name}' is not empty.")
        self.collection_name = collection_name


@retry(
    stop=stop_after_attempt(10),
    wait=wait_fixed(0.5),
    retry=retry_if_exception_type(CollectionNotEmptyError),
    reraise=True,
)
async def _empty_collections() -> None:
    col: AsyncIOMotorCollection
    for col in _collections.values():
        count = await col.count_documents({})
        if count > 0:
            raise CollectionNotEmptyError(col.name)



from __future__ import annotations
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Iterable

import pydot
import typer
from ask_shell import new_task, print_to_live
from model_lib import parse_list, parse_model
from rich.tree import Tree
from zero_3rdparty.iter_utils import flat_map

from atlas_init.tf_ext.constants import ATLAS_PROVIDER_NAME
from atlas_init.tf_ext.models import ModuleConfig, ModuleConfigs
from atlas_init.tf_ext.settings import TfDepSettings
from atlas_init.tf_ext.tf_dep import FORCE_INTERNAL_NODES, SKIP_NODES, AtlasGraph, edge_src_dest

logger = logging.getLogger(__name__)


def default_modules() -> list[str]:
    return [
        "mongodbatlas_advanced_cluster",
        "mongodbatlas_cloud_provider_access_authorization",
        "mongodbatlas_project",
        "mongodbatlas_organization",
    ]


def default_skippped_module_resource_types() -> list[str]:
    return [
        "mongodbatlas_cluster",
        "mongodbatlas_flex_cluster",
    ]


def default_module_configs() -> ModuleConfigs:
    return ModuleConfigs(
        root={
            "alerts": ModuleConfig(
                name="Alerts",
                root_resource_types=[
                    "mongodbatlas_alert_configuration",
                ],
            ),
            "auth": ModuleConfig(
                name="Authentication",
                root_resource_types=[
                    "mongodbatlas_api_key",
                    "mongodbatlas_custom_db_role",
                    "mongodbatlas_database_user",
                    "mongodbatlas_project_api_key",
                    "mongodbatlas_team",
                    "mongodbatlas_x509_authentication_database_user",
                ],
                force_include_children=[
                    "mongodbatlas_access_list_api_key",
                ],
            ),
            "ldap": ModuleConfig(
                name="LDAP",
                root_resource_types=[
                    "mongodbatlas_ldap_configuration",
                    "mongodbatlas_ldap_verify",
                ],
            ),
            "federated auth": ModuleConfig(
                name="Federated Authentication",
                root_resource_types=[
                    "mongodbatlas_federated_settings_identity_provider",
                ],
                force_include_children=[
                    "mongodbatlas_federated_settings_org_config",
                ],
            ),
            "federated DB": ModuleConfig(
                name="Federated Database",
                root_resource_types=[
                    "mongodbatlas_federated_database_instance",
                ],
            ),
            "network": ModuleConfig(
                name="Network",
                root_resource_types=[
                    "mongodbatlas_encryption_at_rest_private_endpoint",
                    "mongodbatlas_network_container",
                    "mongodbatlas_private_endpoint_regional_mode",
                    "mongodbatlas_privatelink_endpoint_service_data_federation_online_archive",
                    "mongodbatlas_privatelink_endpoint_service",
                    "mongodbatlas_privatelink_endpoint",
                    "mongodbatlas_stream_privatelink_endpoint",
                ],
                force_include_children=["mongodbatlas_network_peering"],
            ),
            "cloud_provider": ModuleConfig(
                name="Cloud Provider",
                root_resource_types=["mongodbatlas_cloud_provider_access_setup"],
                allow_external_dependencies=True,
                extra_nested_resource_types=["mongodbatlas_cloud_provider_access_authorization"],
                force_include_children=["mongodbatlas_encryption_at_rest"],
            ),
            "streams": ModuleConfig(
                name="Streams",
                root_resource_types=["mongodbatlas_stream_instance"],
                extra_nested_resource_types=["mongodbatlas_stream_connection"],
                force_include_children=["mongodbatlas_stream_processor"],
            ),
            "cluster": ModuleConfig(
                name="Cluster",
                root_resource_types=["mongodbatlas_advanced_cluster"],
            ),
            "project": ModuleConfig(
                name="Project",
                root_resource_types=["mongodbatlas_project"],
                force_include_children=[
                    "mongodbatlas_project_ip_access_list"  # the external aws_vpc dependency is not really needed
                ],
            ),
            "organization": ModuleConfig(
                name="Organization",
                root_resource_types=["mongodbatlas_organization"],
            ),
            "cloud_backup_actions": ModuleConfig(
                name="Cloud Backup Actions",
                root_resource_types=[
                    "mongodbatlas_cloud_backup_snapshot",
                    "mongodbatlas_cloud_backup_snapshot_export_bucket",
                ],
            ),
        }
    )


def tf_modules(
    skipped_module_resource_types: list[str] = typer.Option(
        ...,
        "-s",
        "--skip-resource-types",
        help="List of resource types to skip when creating module graphs",
        default_factory=default_skippped_module_resource_types,
        show_default=True,
    ),
):
    settings = TfDepSettings.from_env()
    atlas_graph = parse_atlas_graph(settings)
    output_dir = settings.static_root
    with new_task("Write graphs"):
        color_coder_internal = color_coder(atlas_graph, keep_provider_name=False)
        internal_graph = create_internal_dependencies(atlas_graph, color_coder=color_coder_internal)
        add_unused_nodes_to_graph(settings, atlas_graph, color_coder_internal, internal_graph)
        write_graph(internal_graph, output_dir, "atlas_internal.png")
        write_graph(create_external_dependencies(atlas_graph), output_dir, "atlas_external.png")
    with new_task("Write module graphs"):
        modules = generate_module_graphs(skipped_module_resource_types, settings, atlas_graph)
    with new_task("Internal Graph with Module Numbers"):
        module_color_coder = ModuleColorCoder(
            atlas_graph,
            keep_provider_name=False,
            modules=modules,
        )
        internal_graph_with_numbers = create_internal_dependencies(atlas_graph, module_color_coder)
        add_unused_nodes_to_graph(settings, atlas_graph, module_color_coder, internal_graph_with_numbers)
        write_graph(internal_graph_with_numbers, settings.static_root, "atlas_internal_with_numbers.png")
    with new_task("Missing modules"):
        all_resources: list[str] = parse_list(settings.schema_resource_types_path, format="yaml")
        missing_resources = [
            resource_type
            for resource_type in all_resources
            if modules.module_emoji_prefix(resource_type) == ""
            and resource_type not in skipped_module_resource_types
            and resource_type not in atlas_graph.deprecated_resource_types
        ]
        logger.info(f"Missing modules: \n{'\n'.join(missing_resources)}")


def generate_module_graphs(skipped_module_resource_types, settings, atlas_graph):
    tree = Tree(
        "Module graphs",
    )
    used_resource_types: set[str] = set(
        skipped_module_resource_types
    )  # avoid the same resource_type in multiple module graphs
    modules = default_module_configs()
    for name, module_config in modules.root.items():
        internal_graph, external_graph = create_module_graphs(
            atlas_graph,
            module_config,
            color_coder_internal=color_coder(atlas_graph, keep_provider_name=False),
            color_coder_external=color_coder(atlas_graph, keep_provider_name=True),
            used_resource_types=used_resource_types,
        )
        module_tree = tree.add(module_config.tree_label)
        module_trees: dict[str, Tree] = {
            resource_type: module_tree.add(remove_provider_name(resource_type))
            for resource_type in module_config.root_resource_types
        }

        def get_tree(resource_type: str) -> Tree | None:
            return next(
                tree
                for src, tree in module_trees.items()
                if src.endswith(resource_type)  # provider name might be removed
            )

        def prefer_root_src_over_nested(src_dest: tuple[str, str]) -> tuple[bool, str, str]:
            src, dest = src_dest
            is_root = any(root.endswith(src) for root in module_config.root_resource_types)
            return (not is_root, src, dest)  # sort by whether src is a

        for src, dest in sorted(
            (edge_src_dest(edge) for edge in internal_graph.get_edge_list()), key=prefer_root_src_over_nested
        ):
            try:
                tree_src = get_tree(src)
            except StopIteration:
                resource_type = next(
                    root
                    for root in module_config.root_resource_types + module_config.extra_nested_resource_types
                    if root.endswith(src)  # provider name might be removed
                )
                tree_src = module_tree.add(src)
                module_trees[resource_type] = tree_src
            assert tree_src is not None, f"Source {src} not found in module tree"
            module_trees[dest] = tree_src.add(dest)
        write_graph(internal_graph, settings.static_root, f"{name}_internal.png")
        write_graph(external_graph, settings.static_root, f"{name}_external.png")
    print_to_live(tree)
    return modules


def parse_atlas_graph(settings: TfDepSettings) -> AtlasGraph:
    atlas_graph = parse_model(settings.atlas_graph_path, t=AtlasGraph)
    deprecated_resources = parse_list(settings.schema_resource_types_deprecated_path, format="yaml")
    atlas_graph.deprecated_resource_types.update(deprecated_resources)
    atlas_graph.parent_child_edges["mongodbatlas_project"].add("mongodbatlas_auditing")
    atlas_graph.parent_child_edges["mongodbatlas_project"].add("mongodbatlas_custom_dns_configuration_cluster_aws")
    atlas_graph.parent_child_edges["mongodbatlas_advanced_cluster"].add("mongodbatlas_global_cluster_config")
    return atlas_graph


def add_unused_nodes_to_graph(
    settings: TfDepSettings, atlas_graph: AtlasGraph, color_coder: ColorCoder, internal_graph: pydot.Dot
):
    schema_resource_types: list[str] = parse_list(settings.schema_resource_types_path, format="yaml")
    all_nodes = atlas_graph.all_internal_nodes
    for resource_type in schema_resource_types:
        if resource_type not in all_nodes:
            internal_graph.add_node(color_coder.create_node(resource_type, is_unused=True))


class NodeSkippedError(Exception):
    """Raised when a node is skipped during graph creation."""

    def __init__(self, resource_type: str):
        self.resource_type = resource_type
        super().__init__(f"Node skipped: {resource_type}. This is expected for some resource types.")


@dataclass
class ColorCoder:
    graph: AtlasGraph
    keep_provider_name: bool

    ATLAS_EXTERNAL_COLOR: ClassVar[str] = "red"
    ATLAS_INTERNAL_COLOR: ClassVar[str] = "green"
    ATLAS_INTERNAL_UNUSED_COLOR: ClassVar[str] = "gray"
    ATLAS_DEPRECATED_COLOR: ClassVar[str] = "orange"
    EXTERNAL_COLOR: ClassVar[str] = "purple"

    def create_node(self, resource_type: str, *, is_unused: bool = False) -> pydot.Node:
        if resource_type in self.graph.deprecated_resource_types:
            color = self.ATLAS_DEPRECATED_COLOR
        elif is_unused:
            color = self.ATLAS_INTERNAL_UNUSED_COLOR
        elif resource_type.startswith(ATLAS_PROVIDER_NAME):
            color = (
                "red"
                if resource_type in self.graph.all_external_nodes and resource_type not in FORCE_INTERNAL_NODES
                else "green"
            )
        else:
            color = self.EXTERNAL_COLOR
        return pydot.Node(self.node_name(resource_type), shape="box", style="filled", fillcolor=color)

    def node_name(self, resource_type: str) -> str:
        if resource_type in SKIP_NODES:
            raise NodeSkippedError(resource_type)
        return resource_type if self.keep_provider_name else remove_provider_name(resource_type)


def color_coder(atlas_graph: AtlasGraph, keep_provider_name: bool = False) -> ColorCoder:
    return ColorCoder(atlas_graph, keep_provider_name=keep_provider_name)


@dataclass
class ModuleColorCoder(ColorCoder):
    modules: ModuleConfigs

    def node_name(self, resource_type: str) -> str:
        if emoji_prefix := self.modules.module_emoji_prefix(resource_type):
            return f"{emoji_prefix} {super().node_name(resource_type)}"
        return super().node_name(resource_type)


def remove_provider_name(resource_type: str) -> str:
    return resource_type.split("_", 1)[-1]


def write_graph(dot_graph: pydot.Dot, out_path: Path, filename: str):
    out_path.mkdir(parents=True, exist_ok=True)
    dot_graph.write_png(out_path / filename)  # type: ignore


def as_nodes(edges: Iterable[tuple[str, str]]) -> set[str]:
    return set(flat_map((parent, child) for parent, child in edges))


def create_dot_graph(name: str, edges: Iterable[tuple[str, str]], *, color_coder: ColorCoder) -> pydot.Dot:
    edges = sorted(edges)
    graph = pydot.Dot(name, graph_type="graph")
    nodes = as_nodes(edges)
    for node in nodes:
        try:
            graph.add_node(color_coder.create_node(node))
        except NodeSkippedError:
            continue
    for src, dst in edges:
        try:
            graph.add_edge(pydot.Edge(color_coder.node_name(src), color_coder.node_name(dst), color="blue"))
        except NodeSkippedError:
            continue
    return graph


def create_module_graphs(
    atlas_graph: AtlasGraph,
    module_config: ModuleConfig,
    *,
    color_coder_internal: ColorCoder,
    color_coder_external: ColorCoder,
    used_resource_types: set[str],
) -> tuple[pydot.Dot, pydot.Dot]:
    used_resource_types = used_resource_types or set()
    """Create two graphs: one for internal-only module dependencies and one for all module dependencies."""
    child_edges = [
        (root_resource_type, child)
        for root_resource_type in module_config.root_resource_types
        for child in atlas_graph.parent_child_edges.get(root_resource_type, [])
        if child not in used_resource_types
    ]
    child_edges.extend(
        (nested_resource_type, child)
        for nested_resource_type in module_config.extra_nested_resource_types
        for child in atlas_graph.parent_child_edges.get(nested_resource_type, [])
        if child not in used_resource_types
    )
    internal_only_edges = [
        (resource_type, child)
        for resource_type, child in child_edges
        if module_config.include_child(child, atlas_graph)
    ]
    module_name = module_config.name
    internal_graph = create_dot_graph(
        f"{module_name} Internal Only Dependencies",
        internal_only_edges,
        color_coder=color_coder_internal,
    )
    external_edges = [
        (parent, child)
        for child, parents in atlas_graph.external_parents.items()
        if child in child_edges
        for parent in parents
    ]
    external_graph = create_dot_graph(
        f"{module_name} External Dependencies",
        child_edges + external_edges,
        color_coder=color_coder_external,
    )
    used_resource_types.update(module_config.root_resource_types)  # in case a root_resource_type doesn't have children
    used_resource_types |= as_nodes(internal_only_edges)
    return internal_graph, external_graph


def create_internal_dependencies(atlas_graph: AtlasGraph, color_coder: ColorCoder) -> pydot.Dot:
    graph_name = "Atlas Internal Dependencies"
    return create_dot_graph(graph_name, atlas_graph.iterate_internal_edges(), color_coder=color_coder)


def create_external_dependencies(atlas_graph: AtlasGraph) -> pydot.Dot:
    graph_name = "Atlas External Dependencies"
    return create_dot_graph(
        graph_name, atlas_graph.iterate_external_edges(), color_coder=color_coder(atlas_graph, keep_provider_name=True)
    )
